#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
from typing import Any, Optional

from splunk_add_on_ucc_framework import global_config as global_config_lib, utils
from splunk_add_on_ucc_framework.entity import (
    collapse_entity,
    IntervalEntity,
)
from splunk_add_on_ucc_framework.global_config import GlobalConfig
from splunk_add_on_ucc_framework.tabs import resolve_tab

logger = logging.getLogger("ucc_gen")


def _version_tuple(version: str) -> tuple[str, ...]:
    """
    Convert string into tuple to compare versions.

    Args:
        version: raw string
    Returns:
        tuple: version into tuple format
    """
    filled = []
    for point in version.split("."):
        filled.append(point.zfill(8))
    return tuple(filled)


def _handle_biased_terms(conf_entities: list[dict[str, Any]]) -> list[dict[str, Any]]:
    for entity in conf_entities:
        entity_option = entity.get("options")
        if entity_option and "whiteList" in entity_option:
            entity_option["allowList"] = entity_option.get("whiteList")
            del entity_option["whiteList"]
        if entity_option and "blackList" in entity_option:
            entity_option["denyList"] = entity_option.get("blackList")
            del entity_option["blackList"]
    return conf_entities


def _handle_biased_terms_update(global_config: global_config_lib.GlobalConfig) -> None:
    for tab in global_config.configuration:
        conf_entities = tab.get("entity")

        if conf_entities is None:
            continue

        tab["entity"] = _handle_biased_terms(conf_entities)
    for service in global_config.inputs:
        conf_entities = service.get("entity")
        service["entity"] = _handle_biased_terms(conf_entities)
    global_config.update_schema_version("0.0.1")


def _handle_dropping_api_version_update(
    global_config: global_config_lib.GlobalConfig,
) -> None:
    if global_config.meta.get("apiVersion"):
        del global_config.meta["apiVersion"]
    global_config.update_schema_version("0.0.3")


def _handle_alert_action_updates(
    global_config: global_config_lib.GlobalConfig, global_config_path: str
) -> None:
    if global_config.has_alerts():
        updated_alerts = []
        for alert in global_config.alerts:
            modified_alert = {}
            for k, v in alert.items():
                if k in ["activeResponse", "adaptiveResponse"]:
                    # set default values for the below properties
                    v["supportsAdhoc"] = v.get("supportsAdhoc", False)
                    v["supportsCloud"] = v.get("supportsCloud", True)
                if k == "activeResponse":
                    logger.warning(
                        "'activeResponse' is deprecated. Please use 'adaptiveResponse' instead."
                    )
                    modified_alert["adaptiveResponse"] = v
                else:
                    modified_alert[k] = v

            # in either case, we create a new list and fill it with updated alerts, if any
            updated_alerts.append(modified_alert)
        global_config._content["alerts"] = updated_alerts
        global_config.dump(global_config_path)
    global_config.update_schema_version("0.0.4")


def _handle_xml_dashboard_update(
    global_config: global_config_lib.GlobalConfig, global_config_path: str
) -> None:
    panels_to_migrate = [
        "addon_version",
        "events_ingested_by_sourcetype",
        "errors_in_the_addon",
    ]
    if global_config.has_dashboard():
        panels = [panel["name"] for panel in global_config.dashboard["panels"]]
        deprecated_panels = [el for el in panels if el in panels_to_migrate]
        if deprecated_panels:
            logger.warning(
                f"deprecated dashboard panels found: {deprecated_panels}. "
                f"Instead, use just one panel: \"'name': 'default'\""
            )
            global_config.dashboard["panels"] = [{"name": "default"}]
            global_config.dump(global_config_path)
    global_config.update_schema_version("0.0.5")


def handle_global_config_update(
    global_config: global_config_lib.GlobalConfig, global_config_path: str
) -> None:
    """Handle changes in globalConfig file."""
    version = global_config.schema_version or "0.0.0"
    logger.info(f"Current globalConfig schema version is {version}")

    allowed_versions_of_schema_version = {
        "0.0.0",
        "0.0.1",
        "0.0.2",
        "0.0.3",
        "0.0.4",
        "0.0.5",
        "0.0.6",
        "0.0.7",
        "0.0.8",
        "0.0.9",
    }

    if version not in allowed_versions_of_schema_version:
        logger.warning(
            "Schema version is not in the allowed versions, setting it to 0.0.0"
        )
        version = "0.0.0"

    if _version_tuple(version) < _version_tuple("0.0.1"):
        _handle_biased_terms_update(global_config)
        global_config.dump(global_config_path)
        logger.info("Updated globalConfig schema to version 0.0.1")

    if _version_tuple(version) < _version_tuple("0.0.2"):
        for tab in global_config.configuration:
            if tab.get("type") in ["loggingTab", "proxyTab"]:
                continue
            if tab["name"] == "account":
                conf_entities = tab.get("entity")

                if conf_entities is None:
                    continue

                oauth_state_enabled_entity = {}
                for entity in conf_entities:
                    if entity.get("field") == "oauth_state_enabled":
                        logger.warning(
                            "oauth_state_enabled field is no longer a separate "
                            "entity since UCC version 5.0.0. It is now an "
                            "option in the oauth field. Please update the "
                            "globalConfig file accordingly."
                        )
                        oauth_state_enabled_entity = entity

                    if entity.get("field") == "oauth" and not entity.get(
                        "options", {}
                    ).get("oauth_state_enabled"):
                        entity["options"]["oauth_state_enabled"] = False

                if oauth_state_enabled_entity:
                    conf_entities.remove(oauth_state_enabled_entity)

            tab_options = tab.get("options", {})
            if tab_options.get("onChange"):
                logger.error(
                    "The onChange option is no longer supported since UCC "
                    "version 5.0.0. You can use custom hooks to implement "
                    "these actions."
                )
                del tab_options["onChange"]
            if tab_options.get("onLoad"):
                logger.error(
                    "The onLoad option is no longer supported since UCC "
                    "version 5.0.0. You can use custom hooks to implement "
                    "these actions."
                )
                del tab_options["onLoad"]

        if global_config.has_inputs():
            for service in global_config.inputs:
                service_options = service.get("options", {})
                if service_options.get("onChange"):
                    logger.error(
                        "The onChange option is no longer supported since UCC "
                        "version 5.0.0. You can use custom hooks to implement "
                        "these actions."
                    )
                    del service_options["onChange"]
                if service_options.get("onLoad"):
                    logger.error(
                        "The onLoad option is no longer supported since UCC "
                        "version 5.0.0. You can use custom hooks to implement "
                        "these actions."
                    )
                    del service_options["onLoad"]
        global_config.update_schema_version("0.0.2")
        global_config.dump(global_config_path)
        logger.info("Updated globalConfig schema to version 0.0.2")

    if _version_tuple(version) < _version_tuple("0.0.3"):
        _handle_dropping_api_version_update(global_config)
        global_config.dump(global_config_path)
        logger.info("Updated globalConfig schema to version 0.0.3")

    if _version_tuple(version) < _version_tuple("0.0.4"):
        _handle_alert_action_updates(global_config, global_config_path)
        global_config.dump(global_config_path)
        logger.info("Updated globalConfig schema to version 0.0.4")

    if _version_tuple(version) < _version_tuple("0.0.5"):
        _handle_xml_dashboard_update(global_config, global_config_path)
        global_config.dump(global_config_path)
        logger.info("Updated globalConfig schema to version 0.0.5")

    if _version_tuple(version) < _version_tuple("0.0.6"):
        global_config.update_schema_version("0.0.6")
        _dump_with_migrated_tabs(global_config, global_config_path)
        logger.info("Updated globalConfig schema to version 0.0.6")

    if _version_tuple(version) < _version_tuple("0.0.7"):
        global_config.update_schema_version("0.0.7")
        _dump_with_migrated_entities(
            global_config, global_config_path, [IntervalEntity]
        )
        logger.info("Updated globalConfig schema to version 0.0.7")

    if _version_tuple(version) < _version_tuple("0.0.8"):
        _stop_build_on_placeholder_usage(global_config)
        global_config.dump(global_config_path)
        logger.info("Updated globalConfig schema to version 0.0.8")

    if _version_tuple(version) < _version_tuple("0.0.9"):
        _dump_enable_from_global_config(global_config)
        global_config.dump(global_config_path)
        logger.info("Updated globalConfig schema to version 0.0.9")

    if _version_tuple(version) < _version_tuple("0.0.10"):
        _remove_oauth_field_from_entites(global_config)
        global_config.dump(global_config_path)
        logger.info("Updated globalConfig schema to version 0.0.10")


def _dump_with_migrated_tabs(global_config: GlobalConfig, path: str) -> None:
    for i, tab in enumerate(
        global_config.content.get("pages", {}).get("configuration", {}).get("tabs", [])
    ):
        if tab.get("type") == "proxyTab":
            # Collapsing the tab is not required for ProxyTab because we can't be certain
            # that a particular tab is a Proxy tab, as we can with a Logging tab.
            continue
        global_config.content["pages"]["configuration"]["tabs"][i] = _collapse_tab(tab)

    _dump(global_config.content, path, global_config.is_yaml)


def _dump_with_migrated_entities(
    global_config: GlobalConfig,
    path: str,
    entity_type: list[Any],
) -> None:
    _collapse_entities(
        global_config.content.get("pages", {}).get("inputs", {}).get("services"),
        entity_type,
    )
    _collapse_entities(
        global_config.content.get("pages", {}).get("configuration", {}).get("tabs"),
        entity_type,
    )
    _dump(global_config.content, path, global_config.is_yaml)


def _collapse_entities(
    items: Optional[list[dict[Any, Any]]],
    entity_type: list[Any],
) -> None:
    if items is None:
        return

    for item in items:
        for i, entity in enumerate(item.get("entity", [])):
            item["entity"][i] = collapse_entity(entity, entity_type)


def _dump(content: dict[Any, Any], path: str, is_yaml: bool) -> None:
    if is_yaml:
        utils.dump_yaml_config(content, path)
    else:
        utils.dump_json_config(content, path)


def _collapse_tab(tab: dict[str, Any]) -> dict[str, Any]:
    return resolve_tab(tab).short_form()


def _stop_build_on_placeholder_usage(
    global_config: global_config_lib.GlobalConfig,
) -> None:
    # we skip this step as we are stopping while validating it.
    # The step is present only for backward compatibility.
    global_config.update_schema_version("0.0.8")


def _dump_enable_from_global_config(
    global_config: global_config_lib.GlobalConfig,
) -> None:
    if global_config.has_inputs():
        # Fetch the table object from global_config
        table = global_config.content.get("pages", {}).get("inputs", {}).get("table")

        if table:  # If the table exists
            actions = table.get("actions", [])
            if "enable" in actions:
                logger.warning(
                    "`enable` attribute found in input's page table action."
                    + f" Removing 'enable' from actions: {actions}"
                )
                actions.remove("enable")
                table["actions"] = actions  # Update the actions in the global_config

        else:  # If no table present, loop through services in inputs
            services = (
                global_config.content.get("pages", {})
                .get("inputs", {})
                .get("services", [])
            )
            for service in services:
                service_table = service.get("table", {})
                actions = service_table.get("actions", [])
                if "enable" in actions:
                    logger.warning(
                        f"`enable` attribute found in service {service.get('name')}'s table action."
                        + f" Removing 'enable' from actions in service: {actions}"
                    )
                    actions.remove("enable")
                    service_table["actions"] = (
                        actions  # Update the actions in the service's table
                    )

    global_config.update_schema_version("0.0.9")


def _remove_oauth_field_from_oauth_entity(
    entity: dict[str, Any],
) -> None:
    """
    Remove the 'oauth_field' field from the oauth entity in globalConfig.
    This is to ensure that the 'oauth_field' field is not used in entities anymore.
    """
    if entity and "field" in entity and "oauth_field" in entity:
        logger.warning(
            "The 'oauth_field' field is no longer supported in entities. "
            "Please remove it from your globalConfig. Field: %s",
            entity["field"],
        )
        del entity["oauth_field"]


def _handle_removal_of_oauth_field_from_entities(
    services_or_tabs: list[dict[str, Any]],
) -> None:
    """
    Remove the 'oauth' field from entities in globalConfig.
    This is to ensure that the 'oauth' field is not used in entities anymore.
    """
    for tab in services_or_tabs:
        conf_entities = tab.get("entity")
        if conf_entities is None:
            continue

        for entity in conf_entities:
            if entity.get("field") == "oauth":
                auth_types = entity["options"].get("auth_type", [])

                for auth in auth_types:
                    auth_entities = entity["options"].get(auth, [])
                    for auth_entity in auth_entities:
                        _remove_oauth_field_from_oauth_entity(auth_entity)


def _remove_oauth_field_from_entites(
    global_config: global_config_lib.GlobalConfig,
) -> None:
    """
    Remove the 'oauth' field from entities in globalConfig.
    This is to ensure that the 'oauth' field is not used in entities anymore.
    """

    _handle_removal_of_oauth_field_from_entities(global_config.inputs)
    _handle_removal_of_oauth_field_from_entities(global_config.configuration)

    global_config.update_schema_version("0.0.10")



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import logging
import shutil
from os import listdir, makedirs, path, remove, sep
from os.path import basename as bn
from os.path import dirname, exists, isdir, join, isfile, abspath
from splunk_add_on_ucc_framework.app_manifest import AppManifest
from typing import Any
import sys

import addonfactory_splunk_conf_parser_lib as conf_parser
import dunamai
import jinja2
import yaml

from splunk_add_on_ucc_framework import exceptions
from splunk_add_on_ucc_framework import app_manifest as app_manifest_lib
from defusedxml import minidom

logger = logging.getLogger("ucc_gen")


def pretty_print_xml(string: str) -> str:
    """
    Returns a pretty-printed XML as a string.
    """
    return minidom.parseString(string).toprettyxml(indent="    ")


def get_j2_env() -> jinja2.Environment:
    # nosemgrep: splunk.autoescape-disabled, python.jinja2.security.audit.autoescape-disabled.autoescape-disabled
    return jinja2.Environment(
        loader=jinja2.FileSystemLoader(join(dirname(__file__), "templates"))
    )


def get_license_path(file_name: str) -> str:
    return join(dirname(__file__), "templates", "Licenses", f"{file_name}.txt")


def get_icons_path(name: str) -> str:
    return join(dirname(__file__), "templates", "Icons", f"{name}")


def check_author_name(source: str, app_manifest: AppManifest) -> None:
    check_path = join(source, "default", "app.conf")
    if isfile(check_path):
        app_conf = conf_parser.TABConfigParser()
        app_conf.read(check_path)
        app_conf_content = app_conf.item_dict()
        if (
            app_manifest.get_authors()[0]["name"]
            != app_conf_content["launcher"]["author"]
        ):
            logger.warning(
                "Conflicting author names are identified between app.manifest and app.conf in the source directory. "
                "Please specify the author name in app.manifest."
            )


def get_app_manifest(source: str) -> app_manifest_lib.AppManifest:
    app_manifest_path = abspath(
        join(source, app_manifest_lib.APP_MANIFEST_FILE_NAME),
    )
    with open(app_manifest_path) as manifest_file:
        app_manifest_content = manifest_file.read()
    try:
        app_manifest = app_manifest_lib.AppManifest(app_manifest_content)
        app_manifest.validate()
        return app_manifest
    except app_manifest_lib.AppManifestFormatException as e:
        logger.error(
            f"Manifest file @ {app_manifest_path} has invalid format.\n"
            f"Please refer to {app_manifest_lib.APP_MANIFEST_WEBSITE}.\n"
            f"Error message: {e}.\n"
        )
        sys.exit(1)


def recursive_overwrite(src: str, dest: str, has_dashboard: bool = True) -> None:
    """
    Method to copy from src to dest recursively.

    Args:
        src (str): Source of copy
        dest (str): Destination to copy
    """
    # TODO: move to shutil.copytree("src", "dst", dirs_exist_ok=True) when Python 3.8+.
    if isdir(src):
        if not isdir(dest):
            makedirs(dest)
        files = listdir(src)
        for f in files:
            recursive_overwrite(join(src, f), join(dest, f), has_dashboard)
    else:
        if exists(dest):
            remove(dest)

        # EnterpriseViewOnlyPreset is the biggest UI dashboard library file
        # that is not used if dashbaord is not present.
        if has_dashboard or "Dashboard." not in dest:
            shutil.copy(src, dest)


def get_os_path(path: str) -> str:
    """
    Returns a path which will be os compatible.

    Args:
        path (str): Path in string

    Return:
        string: Path which will be os compatible.
    """

    if "\\\\" in path:
        path = path.replace("\\\\", sep)
    else:
        path = path.replace("\\", sep)
    path = path.replace("/", sep)
    return path.strip(sep)


def dump_json_config(config: dict[Any, Any], file_path: str) -> None:
    with open(file_path, "w") as f:
        json.dump(config, f, ensure_ascii=False, indent=4)
        f.write("\n")


def dump_yaml_config(config: dict[Any, Any], file_path: str) -> None:
    with open(file_path, "w") as f:
        yaml.dump(config, f, indent=4, sort_keys=False)


def get_version_from_git() -> str:
    try:
        version = dunamai.Version.from_git()
    except RuntimeError:
        raise exceptions.IsNotAGitRepo()
    if not version.stage:
        stage = "+"
    else:
        stage = version.stage[:1]
    try:
        version.serialize(metadata=True, style=dunamai.Style.SemVer)
    except ValueError:
        raise exceptions.CouldNotVersionFromGitException()
    return f"{version.base}{stage}{version.commit}"


def merge_conf_file(
    src_file: str, dst_file: str, merge_mode: str = "stanza_overwrite"
) -> None:
    merge_deny_list = ["default.meta", "README.txt"]
    if bn(src_file) in merge_deny_list:
        return

    sparser = conf_parser.TABConfigParser()
    sparser.read(src_file)
    src_dict = sparser.item_dict()
    parser = conf_parser.TABConfigParser()
    parser.read(dst_file)
    dst_dict = parser.item_dict()

    if merge_mode == "stanza_overwrite":
        for stanza, key_values in src_dict.items():
            if stanza not in dst_dict:
                parser.add_section(stanza)
            else:
                parser.remove_section(stanza)
                parser.add_section(stanza)

            for k, v in key_values.items():
                parser.set(stanza, k, v)
    elif merge_mode == "item_overwrite":
        for stanza, key_values in src_dict.items():
            if stanza not in dst_dict:
                parser.add_section(stanza)

            for k, v in key_values.items():
                if v:
                    parser.set(stanza, k, v)
                else:
                    parser.remove_option(stanza, k)
    else:
        # overwrite the whole file
        parser.read(src_file)

    with open(dst_file, "w") as df:
        parser.write(df)


def write_file(file_name: str, file_path: str, content: str, **kwargs: str) -> None:
    """
    :param merge_mode: only supported for .conf and .conf.spec files.
    """
    logger.debug('operation="write", object="%s" object_type="file"', file_path)

    merge_mode = kwargs.get("merge_mode", "stanza_overwrite")
    do_merge = False
    if file_name.endswith(".conf") or file_name.endswith(".conf.spec"):
        do_merge = True
    else:
        logger.debug(
            f"'{file_name}' is not going to be merged, only .conf and "
            f".conf.spec files are supported."
        )

    new_file = None
    if path.exists(file_path) and do_merge:
        new_file = path.join(path.dirname(file_path), "new_" + file_name)
    if new_file:
        try:
            with open(new_file, "w+") as fhandler:
                fhandler.write(content)
            merge_conf_file(new_file, file_path, merge_mode=merge_mode)
        finally:
            if path.exists(new_file):
                remove(new_file)
    else:
        if not path.exists(path.dirname(file_path)):
            makedirs(path.dirname(file_path))
        with open(file_path, "w+") as fhandler:
            fhandler.write(content)
        if do_merge:
            parser = conf_parser.TABConfigParser()
            parser.read(file_path)
            with open(file_path, "w") as df:
                parser.write(df)



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional

from splunk_add_on_ucc_framework.generators.file_generator import FileGenerator
from splunk_add_on_ucc_framework.global_config import GlobalConfig


class SearchbnfConf(FileGenerator):
    __description__ = "Generates `searchbnf.conf` for custom search commands provided in the globalConfig."

    def __init__(
        self, global_config: GlobalConfig, input_dir: str, output_dir: str
    ) -> None:
        super().__init__(global_config, input_dir, output_dir)
        self.conf_file = "searchbnf.conf"
        self.searchbnf_info = []
        if global_config.has_custom_search_commands():
            for command in global_config.custom_search_commands:
                if command.get("requiredSearchAssistant", False):
                    searchbnf_dict = {
                        "command_name": command["commandName"],
                        "description": command["description"],
                        "syntax": command["syntax"],
                        "usage": command["usage"],
                    }
                    self.searchbnf_info.append(searchbnf_dict)

    def generate(self) -> Optional[list[dict[str, str]]]:
        if not self.searchbnf_info:
            return None

        file_path = self.get_file_output_path(["default", self.conf_file])
        self.set_template_and_render(
            template_file_path=["conf_files"], file_name="searchbnf_conf.template"
        )
        rendered_content = self._template.render(
            searchbnf_info=self.searchbnf_info,
        )
        return [
            {
                "file_name": self.conf_file,
                "file_path": file_path,
                "content": rendered_content,
            }
        ]



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import sys
from typing import Any, Optional
from splunk_add_on_ucc_framework.tabs.tab import Tab


logger = logging.getLogger("ucc_gen")

# defaults
NAME = "proxy"
TITLE = "Proxy"

DEFAULT_PROXY_ENABLE = {"type": "checkbox", "label": "Enable", "field": "proxy_enabled"}

DEFAULT_PROXY_TYPE = {
    "type": "singleSelect",
    "label": "Proxy Type",
    "required": False,
    "options": {
        "disableSearch": True,
        "autoCompleteFields": [
            {"value": "http", "label": "http"},
            {"value": "socks4", "label": "socks4"},
            {"value": "socks5", "label": "socks5"},
        ],
    },
    "defaultValue": "http",
    "field": "proxy_type",
}

DEFAULT_HOST = {
    "type": "text",
    "label": "Host",
    "validators": [
        {
            "type": "string",
            "errorMsg": "Max host length is 4096",
            "minLength": 1,
            "maxLength": 4096,
        },
        {
            "type": "regex",
            "errorMsg": "Proxy Host should not have special characters",
            "pattern": "^[a-zA-Z]\\w*$",
        },
    ],
    "field": "proxy_url",
    "required": True,
}

DEFAULT_PORT = {
    "type": "text",
    "label": "Port",
    "validators": [{"type": "number", "range": [1, 65535], "isInteger": True}],
    "field": "proxy_port",
    "required": True,
}

DEFAULT_USERNAME = {
    "type": "text",
    "label": "Username",
    "validators": [
        {
            "type": "string",
            "errorMsg": "Max length of username is 50",
            "minLength": 0,
            "maxLength": 50,
        }
    ],
    "field": "proxy_username",
}

DEFAULT_PASSWORD = {
    "type": "text",
    "label": "Password",
    "validators": [
        {
            "type": "string",
            "errorMsg": "Max length of password is 8192",
            "minLength": 0,
            "maxLength": 8192,
        }
    ],
    "encrypted": True,
    "field": "proxy_password",
}

DEFAULT_DNS_RESOLUTION = {
    "type": "checkbox",
    "label": "DNS resolution",
    "field": "proxy_rdns",
}


class ProxyTab(Tab):
    @property
    def tab_type(self) -> Optional[str]:
        return "proxyTab"

    @classmethod
    def from_definition(cls, definition: dict[str, Any]) -> Optional["Tab"]:
        """
        This function checks if the definition has type == ProxyTab; if it does, it gets converted;
        otherwise, it returns None.
        """
        if definition.get("type") == "proxyTab":
            entity = []
            entity_key_const_dict: dict[str, dict[str, Any]] = {
                "enable_proxy": DEFAULT_PROXY_ENABLE,
                "proxy_type": DEFAULT_PROXY_TYPE,
                "host": DEFAULT_HOST,
                "port": DEFAULT_PORT,
                "username": DEFAULT_USERNAME,
                "password": DEFAULT_PASSWORD,
                "dns_resolution": DEFAULT_DNS_RESOLUTION,
            }

            if ("username" not in definition) ^ ("password" not in definition):
                logger.error("Either of username or password is not mentioned.")
                sys.exit(1)

            elif definition.get("username") != definition.get("password"):
                if (
                    isinstance(definition.get("username"), dict)
                    and definition.get("password") is False
                ):
                    logger.error(
                        "You have updated the username but set the password to 'false' which is not allowed "
                        "set `password = True` for default configuration."
                    )
                    sys.exit(1)
                elif (
                    isinstance(definition.get("password"), dict)
                    and definition.get("username") is False
                ):
                    logger.error(
                        "You have updated the password but set username to `false` which is not allowed "
                        "set `username = True` for default configuration."
                    )
                    sys.exit(1)
                elif (
                    type(definition.get("username")) is bool
                    and type(definition.get("password")) is bool
                ):
                    logger.error(
                        f"You have set different values for username ({definition.get('username')})"
                        f" and password ({definition.get('password')}). They should be same."
                    )
                    sys.exit(1)
            for key_name, value in entity_key_const_dict.items():
                if definition.get(key_name) is True:
                    entity.append(value)
                elif (not definition.get(key_name)) and key_name in [
                    "proxy_type",
                    "username",
                    "password",
                    "dns_resolution",
                ]:
                    continue
                elif definition.get(key_name):
                    for key, i_value in value.items():
                        if key not in definition.get(key_name, {}):
                            definition[key_name][key] = i_value
                    entity.append(definition[key_name])
                else:
                    entity.append(value)

            new_definition = {
                "name": definition.get("name", NAME),
                "title": definition.get("title", TITLE),
                "entity": entity,
            }

            for key, value in definition.items():
                if key not in [
                    "name",
                    "title",
                    "enable_proxy",
                    "proxy_type",
                    "port",
                    "host",
                    "username",
                    "password",
                    "dns_resolution",
                    "type",
                ]:
                    new_definition[key] = value

            return ProxyTab(new_definition)

        return None



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# TODO: Update the list as and when Splunk introduces new commands.
# Links to use: https://docs.splunk.com/Documentation/SplunkCloud/latest/SearchReference
#               https://docs.splunk.com/Documentation/Splunk/latest/SearchReference
SPLUNK_COMMANDS = [
    "abstract",
    "accum",
    "addcoltotals",
    "addinfo",
    "addtotals",
    "analyzefields",
    "anomalies",
    "anomalousvalue",
    "anomalydetection",
    "append",
    "appendcols",
    "appendpipe",
    "arules",
    "associate",
    "autoregress",
    "awssnsalert",
    "bin",
    "bucket",
    "bucketdir",
    "chart",
    "cluster",
    "cofilter",
    "collect",
    "concurrency",
    "contingency",
    "convert",
    "correlate",
    "ctable",
    "datamodel",
    "datamodelsimple",
    "dbinspect",
    "dbxquery",
    "dedup",
    "delete",
    "delta",
    "diff",
    "entitymerge",
    "erex",
    "eval",
    "eventcount",
    "eventstats",
    "extract",
    "fieldformat",
    "fields",
    "fieldsummary",
    "filldown",
    "fillnull",
    "findtypes",
    "folderize",
    "foreach",
    "format",
    "from",
    "fromjson",
    "gauge",
    "gentimes",
    "geom",
    "geomfilter",
    "geostats",
    "head",
    "highlight",
    "history",
    "iconify",
    "inputcsv",
    "inputintelligence",
    "inputlookup",
    "iplocation",
    "join",
    "kmeans",
    "kvform",
    "loadjob",
    "localize",
    "localop",
    "lookup",
    "makecontinuous",
    "makemv",
    "makeresults",
    "map",
    "mcollect",
    "metadata",
    "metasearch",
    "meventcollect",
    "mpreview",
    "msearch",
    "mstats",
    "multikv",
    "multisearch",
    "mvcombine",
    "mvexpand",
    "nomv",
    "outlier",
    "outputcsv",
    "outputlookup",
    "outputtext",
    "overlap",
    "pivot",
    "predict",
    "rangemap",
    "rare",
    "regex",
    "reltime",
    "rename",
    "replace",
    "require",
    "rest",
    "return",
    "reverse",
    "rex",
    "rtorder",
    "run",
    "savedsearch",
    "script",
    "scrub",
    "search",
    "searchtxn",
    "selfjoin",
    "sendalert",
    "sendemail",
    "set",
    "setfields",
    "sichart",
    "sirare",
    "sistats",
    "sitimechart",
    "sitop",
    "sort",
    "spath",
    "stats",
    "strcat",
    "streamstats",
    "table",
    "tags",
    "tail",
    "timechart",
    "timewrap",
    "tojson",
    "top",
    "transaction",
    "transpose",
    "trendline",
    "tscollect",
    "tstats",
    "typeahead",
    "typelearner",
    "typer",
    "union",
    "uniq",
    "untable",
    "walklex",
    "where",
    "x11",
    "xmlkv",
    "xmlunescape",
    "xpath",
    "xyseries",
]



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Any

from splunk_add_on_ucc_framework import global_config as global_config_lib

from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.base import (
    RestEndpointBuilder,
)
from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.datainput import (
    DataInputEndpointBuilder,
    DataInputEntityBuilder,
)
from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.field import (
    RestFieldBuilder,
)
from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.multiple_model import (
    MultipleModelEndpointBuilder,
    MultipleModelEntityBuilder,
)
from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.oauth_model import (
    OAuthModelEndpointBuilder,
)
from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.single_model import (
    SingleModelEndpointBuilder,
    SingleModelEntityBuilder,
    SingleModelEndpointBuilderWithOauth,
)
from splunk_add_on_ucc_framework.commands.rest_builder.validator_builder import (
    ValidatorBuilder,
)

REST_HANDLER_DEFAULT_MODULE = "splunktaucclib.rest_handler.admin_external"
REST_HANDLER_DEFAULT_CLASS = "AdminExternalHandler"


def _is_true(val: Any) -> bool:
    return str(val).strip().upper() in ("1", "TRUE", "T", "Y", "YES")


class GlobalConfigBuilderSchema:
    def __init__(self, global_config: global_config_lib.GlobalConfig):
        self.global_config = global_config
        self._settings_conf_file_names: list[str] = list()
        self._configs_conf_file_names: list[str] = list()
        self._oauth_conf_file_names: list[str] = list()
        self._configuration_endpoints: dict[str, RestEndpointBuilder] = {}
        self._inputs_endpoints: dict[str, RestEndpointBuilder] = {}
        self._parse_builder_schema()
        self._endpoints: dict[str, RestEndpointBuilder] = {}
        self._preserve_endpoint_dict()

    @property
    def product(self) -> str:
        return self.global_config.product

    @property
    def namespace(self) -> str:
        return self.global_config.namespace

    @property
    def settings_conf_file_names(self) -> list[str]:
        return self._settings_conf_file_names

    @property
    def configs_conf_file_names(self) -> list[str]:
        return self._configs_conf_file_names

    @property
    def oauth_conf_file_names(self) -> list[str]:
        return self._oauth_conf_file_names

    @property
    def endpoints(self) -> list[RestEndpointBuilder]:
        return list(self._endpoints.values())

    @property
    def configuration_endpoints(self) -> list[RestEndpointBuilder]:
        return list(self._configuration_endpoints.values())

    @property
    def inputs_endpoints(self) -> list[RestEndpointBuilder]:
        return list(self._inputs_endpoints.values())

    @property
    def need_reload(self) -> bool:
        return False

    def _preserve_endpoint_dict(self) -> dict[str, RestEndpointBuilder]:
        self._endpoints.update(self._configuration_endpoints)
        self._endpoints.update(self._inputs_endpoints)
        return self._endpoints

    def _parse_builder_schema(self) -> None:
        self._builder_configs()
        self._builder_settings()
        self._builder_inputs()

    def _builder_configs(self) -> None:
        oauth_handler = False
        token_endpoint = ""
        auth_condition = True

        for config in self.global_config.configs:
            # If we have given oauth support then we have to add endpoint for access_token
            for entity_element in config["entity"]:
                if entity_element["type"] == "oauth":
                    log_details = self.global_config.logging_tab
                    oauth_endpoint = OAuthModelEndpointBuilder(
                        name="oauth",
                        namespace=self.global_config.namespace,
                        app_name=self.global_config.product,
                        log_stanza=log_details.get("name"),
                        log_level_field=log_details.get("entity", [{}])[0].get("field"),
                    )
                    self._configuration_endpoints["oauth"] = oauth_endpoint
                    if oauth_endpoint.conf_name not in self._oauth_conf_file_names:
                        self._oauth_conf_file_names.append(oauth_endpoint.conf_name)

                    auth_types = entity_element["options"]["auth_type"]

                    if "oauth_client_credentials" in auth_types:
                        oauth_handler = True
                        token_endpoint = entity_element["options"][
                            "access_token_endpoint"
                        ]

                        if len(auth_types) == 1:
                            # If we have only the oauth_client_credentials auth type and nothing else
                            # we will not add the auth_type condition
                            auth_condition = False

            name = config["name"]

            endpoint_params = dict(
                name=name,
                namespace=self.global_config.namespace,
                rest_handler_name=config.get("restHandlerName"),
                rest_handler_module=config.get(
                    "restHandlerModule", REST_HANDLER_DEFAULT_MODULE
                ),
                rest_handler_class=config.get(
                    "restHandlerClass", REST_HANDLER_DEFAULT_CLASS
                ),
                need_reload=self.need_reload,
            )

            if oauth_handler:
                endpoint_params["token_endpoint"] = token_endpoint
                endpoint_params["app_name"] = self.global_config.product
                endpoint_params["auth_condition"] = auth_condition
                endpoint: SingleModelEndpointBuilder = (
                    SingleModelEndpointBuilderWithOauth(**endpoint_params)
                )
            else:
                endpoint = SingleModelEndpointBuilder(**endpoint_params)

            self._configuration_endpoints[name] = endpoint
            content = self._get_oauth_enitities(config["entity"])
            fields, special_fields = self._parse_fields(content)
            entity = SingleModelEntityBuilder(
                None,
                fields,
                special_fields=special_fields,
                conf_name=config.get("conf"),
            )
            endpoint.add_entity(entity)

            if endpoint.conf_name not in self._configs_conf_file_names:
                self._configs_conf_file_names.append(endpoint.conf_name)

    def _builder_settings(self) -> None:
        if not self.global_config.settings:
            return

        endpoint = MultipleModelEndpointBuilder(
            name="settings",
            namespace=self.global_config.namespace,
            rest_handler_module=REST_HANDLER_DEFAULT_MODULE,
            rest_handler_class=REST_HANDLER_DEFAULT_CLASS,
            need_reload=self.need_reload,
        )
        self._configuration_endpoints["settings"] = endpoint
        for setting in self.global_config.settings:
            if setting.get("entity") is not None:
                content = self._get_oauth_enitities(setting["entity"])
                fields, special_fields = self._parse_fields(content)
                entity = MultipleModelEntityBuilder(
                    setting["name"],
                    fields,
                    special_fields=special_fields,
                )
                endpoint.add_entity(entity)
            if endpoint.conf_name not in self._settings_conf_file_names:
                self._settings_conf_file_names.append(endpoint.conf_name)

    def _builder_inputs(self) -> None:
        for input_item in self.global_config.inputs:
            rest_handler_name = input_item.get("restHandlerName")
            rest_handler_module = input_item.get(
                "restHandlerModule",
                REST_HANDLER_DEFAULT_MODULE,
            )
            rest_handler_class = input_item.get(
                "restHandlerClass",
                REST_HANDLER_DEFAULT_CLASS,
            )
            if "conf" in input_item:
                name = input_item["name"]
                single_model_endpoint = SingleModelEndpointBuilder(
                    name=name,
                    namespace=self.global_config.namespace,
                    rest_handler_name=rest_handler_name,
                    rest_handler_module=rest_handler_module,
                    rest_handler_class=rest_handler_class,
                    need_reload=self.need_reload,
                )
                self._inputs_endpoints[name] = single_model_endpoint
                content = self._get_oauth_enitities(input_item["entity"])
                fields, special_fields = self._parse_fields(content)
                single_model_entity = SingleModelEntityBuilder(
                    None,
                    fields,
                    special_fields=special_fields,
                    conf_name=input_item["conf"],
                )
                single_model_endpoint.add_entity(single_model_entity)
            else:
                name = input_item["name"]
                data_input_endpoint = DataInputEndpointBuilder(
                    name=name,
                    namespace=self.global_config.namespace,
                    input_type=input_item["name"],
                    rest_handler_name=rest_handler_name,
                    rest_handler_module=rest_handler_module,
                    rest_handler_class=rest_handler_class,
                )
                self._inputs_endpoints[name] = data_input_endpoint
                content = self._get_oauth_enitities(input_item["entity"])
                fields, special_fields = self._parse_fields(content)
                data_input_entity = DataInputEntityBuilder(
                    None,
                    fields,
                    special_fields=special_fields,
                    input_type=input_item["name"],
                )
                data_input_endpoint.add_entity(data_input_entity)

    def _parse_fields(
        self, fields_content: list[dict[str, Any]]
    ) -> tuple[list[RestFieldBuilder], list[RestFieldBuilder]]:
        fields: list[RestFieldBuilder] = []
        special_fields: list[RestFieldBuilder] = []
        if fields_content:
            for field in fields_content:
                rest_field = RestFieldBuilder(
                    field["field"],
                    _is_true(field.get("required")),
                    _is_true(field.get("encrypted")),
                    field.get("defaultValue"),
                    ValidatorBuilder().build(field.get("validators")),
                )

                if field["field"] != "name":
                    fields.append(rest_field)
                else:
                    special_fields.append(rest_field)
        return fields, special_fields

    """
    If the entity contains type oauth then we need to alter the content to generate proper entities to generate
    the rest handler with the oauth fields
    :param content: json content of entity
    :type content: `json`
    """

    def _get_oauth_enitities(
        self, content: list[dict[str, Any]]
    ) -> list[dict[str, Any]]:
        if content:
            for entity_element in content:
                # Check if we have oauth type
                if entity_element["type"] != "oauth":
                    continue

                auth_types = entity_element["options"]["auth_type"]

                for auth_type in auth_types:
                    content = content + entity_element["options"][auth_type]

                if "oauth" in auth_types or "oauth_client_credentials" in auth_types:
                    # Append OAuth fields if there is at least one auth type
                    content = content + [
                        {"field": "access_token", "encrypted": True},
                        {"field": "refresh_token", "encrypted": True},
                        {"field": "instance_url"},
                    ]

                if len(auth_types) > 1:
                    # Append auth_type field if there are multiple auth types
                    content = content + [{"field": "auth_type"}]

                # We will remove the oauth type entity as we have replaced it with all the entity fields
                content.remove(entity_element)
                break

        return content



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from os import walk, path
from os.path import sep
from typing import Any, Optional
from logging import Logger
import addonfactory_splunk_conf_parser_lib as conf_parser


# Some fields are dynamically generated just like `build` for app.conf
# so while comparing them we should remove them from the dict
STANZA_TO_REMOVE = {"app.conf": {"install": ["build"]}}


class CodeGeneratorDiffChecker:
    COMMON_FILES_MESSAGE_PART_1 = (
        "Below are the file(s) that are auto generated by the UCC framework. "
        "The below files can be removed from your repository:"
    )
    COMMON_FILES_MESSAGE_PART_2 = (
        "Please refer UCC framework documentation for the latest "
        "features that allows you to remove the above files."
    )
    DIFFERENT_STANZA_MESSAGE = (
        "Below are the stanza(s) that are not auto generated by the UCC framework. "
        "(Optionally), you can raise feature requests for UCC framework at "
        "'https://github.com/splunk/addonfactory-ucc-generator/issues/new/choose' "
        "with the output mentioned below."
    )

    DIFFERENT_FILES_MESSAGE = (
        "Below are the file(s) with the attributes of the stanza that are not auto generated by the UCC framework. "
        "(Optionally), you can raise feature requests for UCC framework at "
        "'https://github.com/splunk/addonfactory-ucc-generator/issues/new/choose' "
        "with the output mentioned below."
    )

    def __init__(self, src_dir: str, dst_dir: str) -> None:
        self.source_default_directory = f"{src_dir}/default"
        self.target_default_directory = f"{dst_dir}/default"
        self.different_files: dict[str, Any[dict[str, str], list[dict[str, str]]]] = {}
        self.common_files: dict[str, str] = {}
        self.same_stanza: dict[str, str] = {}
        self.diff_stanza: dict[str, str] = {}

    def deduce_gen_and_custom_content(
        self, logger: Logger, ignore_file_list: Optional[list[str]] = None
    ) -> None:
        """
        Deduce that the files have same content or different
        - For the same content, developer can remove it from the repository
        """
        if ignore_file_list is None:
            ignore_file_list = []
        # we add these two files as they are required to be present in source code
        # TODO: try to implement generation of these files from globalConfig
        ignore_file_list.extend(["app.manifest", "README.txt"])

        src_all_files: dict[str, str] = {}
        for root, _, files in walk(self.source_default_directory):
            for file in files:
                src_all_files[file] = sep.join([root, file])

        dest_all_files: dict[str, str] = {}
        for root, _, files in walk(self.target_default_directory):
            for file in files:
                dest_all_files[file] = sep.join([root, file])
        dest_all_files["default.meta"] = sep.join(
            [self.target_default_directory, "metadata"]
        )

        for file_name in dest_all_files.keys():
            if file_name in src_all_files.keys():
                if file_name in ignore_file_list:
                    continue
                if file_name.endswith(("app.conf", "restmap.conf", "web.conf")):
                    self._conf_file_diff_checker(
                        src_all_files[file_name], dest_all_files[file_name]
                    )

        self.print_files(logger)

    def _conf_file_diff_checker(self, src_file: str, target_file: str) -> None:
        """
        Find the difference between the source code and generated code for the
        conf files created in package/default directory
        """
        sparser = conf_parser.TABConfigParser()
        sparser.read(src_file)
        src_dict = sparser.item_dict()
        parser = conf_parser.TABConfigParser()
        parser.read(target_file)
        dst_dict = parser.item_dict()
        file_name = path.basename(src_file)

        # Remove dynamically generated fields before comparing
        if file_name in STANZA_TO_REMOVE.keys():
            for stanza in STANZA_TO_REMOVE[file_name].keys():
                if stanza in src_dict.keys():
                    for key_value in STANZA_TO_REMOVE[file_name][stanza]:
                        src_dict[stanza].pop(key_value, None)
                        dst_dict[stanza].pop(key_value, None)
        if src_dict == dst_dict:
            self.common_files[src_file] = src_file.split(sep=sep)[-1]
        else:
            for stanza, src_kv in src_dict.items():
                dst_kv = dst_dict.get(stanza)

                # Stanza not present in destination that is we do not generate this whole stanza.
                if dst_kv is None:
                    for key, val in src_kv.items():
                        self.diff_stanza[f"{src_file} :: {stanza}"] = src_file.split(
                            sep=sep
                        )[-1]
                    continue

                # Stanza present, check individual key differences
                for key, val in src_kv.items():
                    if key not in dst_kv:
                        self.different_files[f"{src_file}[{stanza}] :: {key}"] = {
                            "repository": val,
                            "output": "",
                        }
                # If stanza dicts are identical, mark them as same stanza
                if src_kv == dst_kv:
                    self.same_stanza[f"{src_file} :: {stanza}"] = src_file.split(
                        sep=sep
                    )[-1]

    def print_files(self, logger: Logger) -> None:
        """
        Print the common and different files in the console
        """
        messages: list[str] = []
        if self.common_files:
            messages.append("-" * 120)
            messages.append(self.COMMON_FILES_MESSAGE_PART_1)
            messages.extend(
                [f"{idx + 1}) {f}" for idx, f in enumerate(self.common_files.keys())]
            )
            messages.append(self.COMMON_FILES_MESSAGE_PART_2)
            messages.append("-" * 120)
            logger.warning("\n\n".join(messages))

        messages.clear()

        # TODO: uncomment the below code when UCC provide functionality of merging .conf files
        # if self.same_stanza:
        #     messages.append("-" * 120)
        #     messages.append(
        #         "Below are the stanzas that can be removed from your source directory as, UCC auto generates it"
        #     )
        #     messages.extend(
        #         [f"{idx + 1}) {f}" for idx, f in enumerate(self.same_stanza.keys())]
        #     )
        #     messages.append("-" * 120)
        #     logger.warning("\n\n".join(messages))
        # messages.clear()

        # if self.diff_stanza:
        #     messages.append("-" * 120)
        #     messages.append(self.DIFFERENT_STANZA_MESSAGE)
        #     messages.extend(
        #         [f"{idx + 1}) {f}" for idx, f in enumerate(self.diff_stanza.keys())]
        #     )
        #     messages.append("-" * 120)
        #     logger.warning("\n\n".join(messages))
        # messages.clear()

        if self.different_files:
            messages.append("+" * 120)
            messages.append(self.DIFFERENT_FILES_MESSAGE)
            file_count = 1
            for k, v in self.different_files.items():
                file_msg: str = ""
                file_msg = f"{file_count}) {k}"
                # v would be of instance of list when we scope XML file for comparison
                if isinstance(v, dict):
                    file_msg += f"\n\tSource: {v.get('repository')}, Generated: {v.get('output')}"
                messages.append(file_msg)
                file_count += 1
            messages.append("+" * 120)
            logger.warning("\n".join(messages))
        messages.clear()



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import subprocess

import logging
import sys
import json

from typing import Any
from splunk_add_on_ucc_framework import app_manifest as ap

logger = logging.getLogger("ucc_gen")


def import_from_aob(addon_name: str) -> None:
    addon_name_directory = os.path.join(os.getcwd(), addon_name)
    if not os.path.isdir(addon_name_directory):
        logger.error(f"No such directory {addon_name_directory}")
        sys.exit(1)
    import_from_aob_script_path = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        "commands",
        "import_from_aob.sh",
    )
    imports_py_path = os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "commands", "imports.py"
    )
    subprocess.call((import_from_aob_script_path, addon_name, imports_py_path))
    addon_ucc_path = addon_name_directory + "_ucc"
    a_path = os.path.join(*[os.getcwd(), addon_ucc_path, "package", "app.manifest"])
    with open(a_path, encoding="utf-8") as f:
        data = json.load(f)

    add_app_manifest_key(
        data, "supportedDeployments", ap.APP_MANIFEST_SUPPORTED_DEPLOYMENTS
    )
    add_app_manifest_key(data, "targetWorkloads", ap.APP_MANIFEST_TARGET_WORKLOADS)
    check_app_manifest_schema_version(data)

    with open(a_path, "w", encoding="utf-8") as f:
        json.dump(data, f, default=list, ensure_ascii=False, indent=4)


def add_app_manifest_key(
    app_manifest_data: dict[Any, Any], key: str, value: frozenset[str]
) -> None:
    if key not in app_manifest_data:
        app_manifest_data.update({key: value})
    elif app_manifest_data[key] is None:
        app_manifest_data.pop(key)
        app_manifest_data.update({key: value})


def check_app_manifest_schema_version(app_manifest_data: dict[Any, Any]) -> None:
    key = "schemaVersion"
    version = app_manifest_data[key]
    if version != ap.APP_MANIFEST_SCHEMA_VERSION:
        app_manifest_data.pop(key)
        app_manifest_data.update({key: ap.APP_MANIFEST_SCHEMA_VERSION})



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional, Any, TYPE_CHECKING

from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.single_model import (
    RestEndpointBuilder,
    RestEntityBuilder,
)

if TYPE_CHECKING:
    from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.field import (
        RestFieldBuilder,
    )


class DataInputEntityBuilder(RestEntityBuilder):
    def __init__(
        self,
        name: Optional[str],
        fields: list["RestFieldBuilder"],
        input_type: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(name, fields, **kwargs)
        self._input_type = input_type

    @property
    def name_spec(self) -> str:
        return f"{self._input_type}://<name>"

    @property
    def name_rh(self) -> str:
        return ""


class DataInputEndpointBuilder(RestEndpointBuilder):
    _rh_template = """
import import_declare_test

from splunktaucclib.rest_handler.endpoint import (
    field,
    validator,
    RestModel,
    DataInputModel,
)
from splunktaucclib.rest_handler import admin_external, util
from {handler_module} import {handler_class}
import logging

util.remove_http_proxy_env_vars()

{entity}


endpoint = DataInputModel(
    '{input_type}',
    model,
)


if __name__ == '__main__':
    logging.getLogger().addHandler(logging.NullHandler())
    admin_external.handle(
        endpoint,
        handler={handler_class},
    )
"""

    def __init__(
        self, name: str, namespace: str, input_type: str, **kwargs: Any
    ) -> None:
        super().__init__(name, namespace, **kwargs)
        self.input_type = input_type

    @property
    def conf_name(self) -> str:
        return "inputs"

    def actions(self) -> list[str]:
        return ["edit", "list", "remove", "create"]

    def generate_rh(self) -> str:
        entity = self._entities[0]
        return self._rh_template.format(
            handler_module=self.rh_module,
            handler_class=self.rh_class,
            entity=entity.generate_rh(),
            input_type=self.input_type,
        )



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import os.path as op

from splunk_add_on_ucc_framework.commands.rest_builder import (
    global_config_builder_schema,
)
from splunk_add_on_ucc_framework.global_config import OSDependentLibraryConfig

__all__ = ["RestBuilder"]

_import_declare_content = """
import os
import sys
import re
from os.path import dirname

ta_name = '{ta_name}'
pattern = re.compile(r'[\\\\/]etc[\\\\/]apps[\\\\/][^\\\\/]+[\\\\/]bin[\\\\/]?$')
new_paths = [path for path in sys.path if not pattern.search(path) or ta_name in path]
new_paths.insert(0, os.path.join(dirname(dirname(__file__)), "lib"))
new_paths.insert(0, os.path.sep.join([os.path.dirname(__file__), ta_name]))
sys.path = new_paths
"""

_import_declare_os_lib_content = """
bindir = os.path.dirname(os.path.realpath(os.path.dirname(__file__)))
libdir = os.path.join(bindir, "lib")
platform = sys.platform
python_version = "".join(str(x) for x in sys.version_info[:2])
"""


def _generate_import_declare_test(
    schema: global_config_builder_schema.GlobalConfigBuilderSchema,
) -> str:
    base_content = _import_declare_content.format(ta_name=schema.product)
    libraries = schema.global_config.os_libraries
    if not libraries:
        return base_content

    base_content += _import_declare_os_lib_content
    os_lib_part = ""

    paths = group_libs_by_python_version_and_platform(libraries)
    for python_version, os_specific_paths in paths.items():
        os_lib_part += f'\nif python_version == "{python_version}":\n'
        for lib_os, targets in os_specific_paths.items():
            if lib_os == "windows":
                os_lib_part += '\tif platform.startswith("win"):\n'
                for target in targets:
                    os_lib_part += get_insert_to_syspath_str(target)
            elif lib_os == "darwin":
                os_lib_part += '\tif platform.startswith("darwin"):\n'
                for target in targets:
                    os_lib_part += get_insert_to_syspath_str(target)
            else:
                os_lib_part += '\tif platform.startswith("linux"):\n'
                for target in targets:
                    os_lib_part += get_insert_to_syspath_str(target)

    return base_content + os_lib_part


def group_libs_by_python_version_and_platform(
    libraries: list[OSDependentLibraryConfig],
) -> dict[str, dict[str, set[str]]]:
    """Returns os specific paths grouped by python version and platform"""
    python_versions = {lib.python_version for lib in libraries}
    os_specific_paths = {}
    for python_version in python_versions:
        os_specific_paths[python_version] = get_paths_to_add(
            [lib for lib in libraries if lib.python_version == python_version]
        )
    return os_specific_paths


def get_paths_to_add(libraries: list[OSDependentLibraryConfig]) -> dict[str, set[str]]:
    result: dict[str, set[str]] = {}
    for library in libraries:
        lib_os = library.os
        target = os.path.normpath(library.target)
        result.setdefault(lib_os, set()).add(target)
    return result


def get_insert_to_syspath_str(target: str) -> str:
    return f'\t\tsys.path.insert(0, os.path.join(libdir, "{target}"))\n'


class _RestBuilderOutput:
    readme = "README"
    default = "default"
    bin = "bin"

    def __init__(self, path: str) -> None:
        self._path = path
        self._root_path = op.abspath(self._path)
        if not op.isdir(self._root_path):
            os.makedirs(self._root_path)
        self._content: dict[str, list[str]] = {}

    def put(self, subpath: str, file_name: str, content: str) -> None:
        path = op.join(self._root_path, subpath)
        if not op.isdir(path):
            os.makedirs(path)
        full_name = op.join(path, file_name)
        if full_name not in self._content:
            self._content[full_name] = []
        self._content[full_name].append(content)

    def save(self) -> None:
        for full_name, contents in list(self._content.items()):
            full_content = "\n\n".join(contents)
            with open(full_name, "w") as f:
                f.writelines(full_content)


class RestBuilder:
    def __init__(
        self,
        schema: global_config_builder_schema.GlobalConfigBuilderSchema,
        output_path: str,
    ) -> None:
        self._schema = schema
        self._output_path = output_path
        self.output = _RestBuilderOutput(self._output_path)

    def add_executable_attribute(self) -> None:
        def _add_executable_attribute(file_path: str) -> None:
            if op.isfile(file_path):
                st = os.stat(file_path)
                os.chmod(file_path, st.st_mode | 0o111)

        bin_path = os.path.join(
            self._output_path,
            self.output.bin,
        )
        files_under_bin = os.listdir(bin_path)
        for file_path in files_under_bin:
            _add_executable_attribute(file_path)

    def build(self) -> None:
        for endpoint in self._schema.endpoints:
            self.output.put(
                self.output.bin,
                endpoint.rh_name + ".py",
                endpoint.generate_rh(),
            )

        self.output.put(
            self.output.bin,
            "import_declare_test.py",
            _generate_import_declare_test(self._schema),
        )
        self.output.save()



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional

from splunk_add_on_ucc_framework.generators.file_generator import FileGenerator
from splunk_add_on_ucc_framework.global_config import GlobalConfig


class AccountConf(FileGenerator):
    __description__ = (
        "Generates `<YOUR_ADDON_NAME>_account.conf.spec` "
        "file for the configuration mentioned in globalConfig"
    )

    def __init__(
        self, global_config: GlobalConfig, input_dir: str, output_dir: str
    ) -> None:
        super().__init__(global_config, input_dir, output_dir)
        self.account_fields: list[tuple[str, list[str]]] = []
        if global_config.has_configuration():
            self.conf_spec_file = global_config.namespace.lower() + "_account.conf.spec"
            for account in global_config.configs:
                # If the endpoint is oauth, which is for getting access_token, conf file entries
                # should not get created (compatibility to previous versions)
                if account["name"] == "oauth":
                    continue
                content = self._gc_schema._get_oauth_enitities(account["entity"])
                fields, _ = self._gc_schema._parse_fields(content)
                self.account_fields.append(
                    ("<name>", [f"{f._name} = " for f in fields])
                )

    def generate(self) -> Optional[list[dict[str, str]]]:
        if not self.account_fields:
            return None

        file_path = self.get_file_output_path(["README", self.conf_spec_file])
        self.set_template_and_render(
            template_file_path=["README"], file_name="account_conf_spec.template"
        )

        rendered_content = self._template.render(account_stanzas=self.account_fields)
        return [
            {
                "file_name": self.conf_spec_file,
                "file_path": file_path,
                "content": rendered_content,
            }
        ]



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import os
import re
from collections import namedtuple
from pathlib import Path
from typing import Optional

logger = logging.getLogger("ucc_gen")
FileUpdater = namedtuple("FileUpdater", ["path_segments", "function"])


_base_html_pattern = re.compile(
    r"<script\s+src=\"\${make_url\(page_path\)}\"\s*>\s*</script>"
)


def _handle_base_html_update(content: str) -> Optional[str]:
    matches = _base_html_pattern.findall(content)

    if not matches:
        return None

    for result in matches:
        content = content.replace(
            result, '<script type="module" src="${make_url(page_path)}"></script>'
        )

    return content


def handle_package_files_update(path: str) -> None:
    files_to_update = [
        FileUpdater(("appserver", "templates", "base.html"), _handle_base_html_update)
    ]

    for file_updater in files_to_update:
        relative_path = os.path.join(*file_updater.path_segments)
        file_path = os.path.join(path, *file_updater.path_segments)

        if not os.path.isfile(file_path):
            continue

        path_obj = Path(file_path)
        original_content = path_obj.read_text()
        output_content = file_updater.function(original_content)

        if output_content is None or original_content == output_content:
            continue

        path_obj.write_text(output_content)
        logger.info(
            f"File '{relative_path}' exists in the package directory and its content needed to be updated by UCC."
        )



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import sys

logger = logging.getLogger("ucc_gen")


def validate(file_path: str) -> None:
    try:
        from splunk_appinspect import main
    except ModuleNotFoundError:
        logger.error(
            "UCC validate dependencies are not installed. Please install them using the command -> "
            "`pip install splunk-add-on-ucc-framework[validate]`."
        )
        sys.exit(1)
    else:
        main.validate([f"{file_path}", "--included-tags", "cloud"])



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional, Any

from splunk_add_on_ucc_framework.entity.entity import Entity

CRON_REGEX = (
    r"^("
    r"(?:-1|\d+(?:\.\d+)?)"  # Non-negative number or -1
    r"|"
    r"(([\*\d{1,2}\,\-\/]+\s){4}[\*\d{1,2}\,\-\/]+)"  # CRON interval
    r")$"
)


class IntervalEntity(Entity):
    def short_form(self) -> dict[str, Any]:
        return dict(self)

    def long_form(self) -> dict[str, Any]:
        definition = {
            "type": "text",
            "field": self["field"],
            "label": self["label"],
            "validators": [
                {
                    "type": "regex",
                    "errorMsg": f"{self['label']} must be either a non-negative number, CRON interval or -1.",
                    "pattern": CRON_REGEX,
                }
            ],
        }

        options = self.get("options", {})
        range_v = options.get("range")

        if range_v is not None:
            min_v, max_v = range_v
            definition["validators"].append(
                {
                    "type": "number",
                    "range": range_v,
                    "errorMsg": f"{self['label']} must be between {min_v} and {max_v}",
                }
            )

        for name in ("help", "required", "defaultValue", "tooltip"):
            if name in self:
                definition[name] = self[name]

        return definition

    @classmethod
    def from_definition(cls, definition: dict[str, Any]) -> Optional[Entity]:
        if definition.get("type") == "interval":
            return cls(definition)

        if definition.get("type") != "text":
            return None

        if str(definition["field"]).lower() != "interval":
            return None

        validators = {i["type"] for i in list(definition.get("validators", []))}

        if validators != {"regex"} and validators != {"regex", "number"}:
            return None

        new_definition = {
            "type": "interval",
            "field": definition["field"],
            "label": definition["label"],
        }

        for name in ("help", "required", "defaultValue", "tooltip"):
            if name in definition:
                new_definition[name] = definition[name]

        if "number" in validators:
            for v in list(definition["validators"]):
                if v["type"] == "number":
                    new_definition["options"] = {"range": v["range"]}

        return cls(new_definition)



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from splunk_add_on_ucc_framework.generators.file_generator import FileGenerator
from typing import Optional
from xml.etree.ElementTree import Element, SubElement, tostring
from splunk_add_on_ucc_framework.utils import pretty_print_xml
from splunk_add_on_ucc_framework.global_config import GlobalConfig


class InputsXml(FileGenerator):
    __description__ = (
        "Generates inputs.xml based on inputs configuration present in globalConfig,"
        " in `default/data/ui/views/inputs.xml` folder"
    )

    def __init__(
        self, global_config: GlobalConfig, input_dir: str, output_dir: str
    ) -> None:
        super().__init__(global_config, input_dir, output_dir)
        if global_config.has_inputs():
            self.inputs_xml_content = self.generate_views_inputs_xml(
                self._addon_name,
            )

    def generate_views_inputs_xml(self, addon_name: str) -> str:
        """
        Generates `default/data/ui/views/inputs.xml` file.
        """
        view = Element(
            "view",
            attrib={
                "template": f"{addon_name}:/templates/base.html",
                "type": "html",
                "isDashboard": "False",
            },
        )
        label = SubElement(view, "label")
        label.text = "Inputs"
        view_as_string = tostring(view, encoding="unicode")
        return pretty_print_xml(view_as_string)

    def generate(self) -> Optional[list[dict[str, str]]]:
        if not self._global_config.has_inputs():
            return None
        file_path = self.get_file_output_path(
            ["default", "data", "ui", "views", "inputs.xml"]
        )
        return [
            {
                "file_name": "inputs.xml",
                "file_path": file_path,
                "content": self.inputs_xml_content,
            }
        ]



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import logging
import os
import sys

from typing import Any, Optional

from splunk_add_on_ucc_framework import global_config as global_config_lib
from splunk_add_on_ucc_framework import utils

logger = logging.getLogger("ucc_gen")

PANEL_ADDON_VERSION = "addon_version"
PANEL_EVENTS_INGESTED_BY_SOURCETYPE = "events_ingested_by_sourcetype"
PANEL_ERRORS_IN_THE_ADDON = "errors_in_the_addon"
PANEL_DEFAULT = "default"
PANEL_CUSTOM = "custom"

SUPPORTED_PANEL_NAMES = frozenset(
    [
        PANEL_DEFAULT,
        PANEL_CUSTOM,
        PANEL_ADDON_VERSION,
        PANEL_EVENTS_INGESTED_BY_SOURCETYPE,
        PANEL_ERRORS_IN_THE_ADDON,
    ]
)
SUPPORTED_PANEL_NAMES_READABLE = ", ".join(SUPPORTED_PANEL_NAMES)

# default sparkline with 0 values as text
DEFAULT_SPARK_LINE = '\\"##__SPARKLINE__##,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0\\"'

# query to fill table cell chart with zero line if no data is available
FILL_DEFAULT_SPARKLINE_AND_VALUE = (
    f"| fillnull value={DEFAULT_SPARK_LINE} sparkevent | fillnull value=0 events "
)

default_definition_json_filename = {
    "overview": "overview_definition.json",
    "data_ingestion_tab": "data_ingestion_tab_definition.json",
    "errors_tab": "errors_tab_definition.json",
    "resources_tab": "resources_tab_definition.json",
    "data_ingestion_modal_definition": "data_ingestion_modal_definition.json",
}

data_ingestion = (
    "index=_internal source=*license_usage.log type=Usage "
    "({determine_by} IN ({lic_usg_condition})) | timechart sum(b) as Usage | "
    'rename Usage as \\"Data volume\\"'
)

data_ingestion_and_events = (
    "index=_internal source=*license_usage.log type=Usage "
    "({determine_by} IN ({lic_usg_condition})) | timechart sum(b) as Usage "
    '| rename Usage as \\"Data volume\\" '
    "| join _time [search index=_internal source=*{addon_name}* action=events_ingested "
    '| timechart sum(n_events) as \\"Number of events\\" ]'
)

errors_count = "index=_internal source=*{addon_name}* log_level IN ({log_lvl}) | timechart count as Errors by exc_l "

# query generate data if there is 0 data in basic query
# | head (${basic_query_token}:job.resultCount$==0)]" is used to check if there is 0 data in basic query
# requires smart sources enabled,
zero_line_search_query = (
    "| append [ gentimes increment=5m [ makeresults "
    "| eval start=strftime( "
    'if(\\"${time_token}.earliest$\\"=\\"now\\"'
    ",now(),"
    'if(match(\\"${time_token}.earliest$\\",\\"^\\\\d+-\\\\d+-\\\\d+(T?\\\\d+:\\\\d+:\\\\d+(\\\\.\\\\d{{3}}Z)?)$\\"),'
    'strptime(\\"${time_token}.earliest$\\", \\"%Y-%m-%dT%H:%M:%S.%N\\")'
    ',relative_time(now(), \\"${time_token}.earliest$\\")'
    ")"
    "), "
    '\\"%m/%d/%Y:%T\\")'
    "| eval end=strftime("
    'if(\\"${time_token}.latest$\\"=\\"now\\",'
    "now(),"
    'if(match(\\"${time_token}.latest$\\",\\"^\\\\d+-\\\\d+-\\\\d+(T?\\\\d+:\\\\d+:\\\\d+(\\\\.\\\\d{{3}}Z)?)$\\"),'
    'strptime(\\"${time_token}.latest$\\", \\"%Y-%m-%dT%H:%M:%S.%N\\") '
    ',relative_time(now(), \\"${time_token}.latest$\\")'
    ")"
    "), "
    '\\"%m/%d/%Y:%T\\")'
    "| return start end] "
    "| eval {value_label} = 0 | fields - endhuman starthuman starttime "
    "| rename endtime as _time | head (${basic_query_token}:job.resultCount$==0)]"
)

events_count = (
    "index=_internal source=*{addon_name}* action=events_ingested | "
    'timechart sum(n_events) as \\"Number of events\\"'
)

table_sourcetype_query = (
    "index=_internal source=*license_usage.log type=Usage ({determine_by} IN ({lic_usg_condition})) "
    "| fillnull value=0 b | stats sparkline(sum(b)) as sparkvolume, sum(b) as Bytes by st "
    "| join type=left st [search index = _internal source=*{addon_name}* action=events_ingested "
    "| stats latest(_time) AS le, sparkline(sum(n_events)) as sparkevent, "
    "sum(n_events) as events by sourcetype_ingested "
    "| rename sourcetype_ingested as st ] "
    f"{FILL_DEFAULT_SPARKLINE_AND_VALUE}"
    '| makemv delim=\\",\\" sparkevent '
    '| eval \\"Last event\\" = strftime(le, \\"%e %b %Y %I:%M%p\\") '
    '| table st, Bytes, sparkvolume, events, sparkevent, \\"Last event\\" '
    '| rename st as \\"Source type\\", Bytes as \\"Data volume\\", events as \\"Number of events\\", '
    'sparkvolume as \\"Volume trendline (Bytes)\\", sparkevent as \\"Event trendline\\"'
)

table_source_query = (
    "index=_internal source=*license_usage.log type=Usage ({determine_by} IN ({lic_usg_condition})) "
    "| fillnull value=0 b | stats sparkline(sum(b)) as sparkvolume, sum(b) as Bytes by s "
    "| join type=left s [search index = _internal source=*{addon_name}* action=events_ingested "
    "| stats latest(_time) AS le, sparkline(sum(n_events)) as sparkevent, "
    "sum(n_events) as events by modular_input_name "
    "| rename modular_input_name as s ] "
    f"{FILL_DEFAULT_SPARKLINE_AND_VALUE}"
    '| makemv delim=\\",\\" sparkevent '
    '| eval \\"Last event\\" = strftime(le, \\"%e %b %Y %I:%M%p\\") '
    '| table s, Bytes, sparkvolume, events, sparkevent, \\"Last event\\" '
    '| rename s as \\"Source\\", Bytes as \\"Data volume\\", events as \\"Number of events\\", '
    'sparkvolume as \\"Volume trendline (Bytes)\\", sparkevent as \\"Event trendline\\"'
)
table_host_query = (
    "index=_internal source=*license_usage.log type=Usage "
    "({determine_by} IN ({lic_usg_condition})) "
    "| fillnull value=0 b | stats sparkline(sum(b)) as sparkvolume, sum(b) as Bytes by h "
    "| table h, Bytes, sparkvolume "
    '| rename h as \\"Host\\", Bytes as \\"Data volume\\", sparkvolume as \\"Volume trendline (Bytes)\\"'
)
table_index_query = (
    "index=_internal source=*license_usage.log type=Usage ({determine_by} IN ({lic_usg_condition})) "
    "| fillnull value=0 b | stats sparkline(sum(b)) as sparkvolume, sum(b) as Bytes by idx "
    "| join type=left idx [search index = _internal source=*{addon_name}* action=events_ingested "
    "| stats latest(_time) AS le, sparkline(sum(n_events)) as sparkevent, "
    "sum(n_events) as events by event_index "
    "| rename event_index as idx ] "
    f"{FILL_DEFAULT_SPARKLINE_AND_VALUE}"
    '| makemv delim=\\",\\" sparkevent '
    '| eval \\"Last event\\" = strftime(le, \\"%e %b %Y %I:%M%p\\") '
    '| table idx, Bytes, sparkvolume, events, sparkevent, \\"Last event\\" '
    '| rename idx as \\"Index\\", Bytes as \\"Data volume\\", events as \\"Number of events\\", '
    'sparkvolume as \\"Volume trendline (Bytes)\\", sparkevent as \\"Event trendline\\"'
)
table_account_query = (
    "index = _internal source=*{addon_name}* action=events_ingested "
    "| fillnull value=0 n_events "
    "| stats latest(_time) as le, sparkline(sum(n_events)) as sparkevent, sum(n_events) as events by event_account "
    '| eval \\"Last event\\" = strftime(le, \\"%e %b %Y %I:%M%p\\") '
    '| table event_account, events, sparkevent, \\"Last event\\" '
    '| rename event_account as \\"Account\\", events as \\"Number of events\\", '
    'sparkevent as \\"Event trendline\\"'
)

table_input_query = (
    '| rest splunk_server=local /services/data/inputs/all | where $eai:acl.app$ = \\"{addon_name}\\" '
    '| eval Active=if(lower(disabled) IN (\\"1\\", \\"true\\", \\"t\\"), \\"no\\", \\"yes\\") '
    '| table title, Active | rename title as \\"event_input\\" | join type=left event_input [ '
    "search index = _internal source=*{addon_name_lowercase}* action=events_ingested "
    "| stats latest(_time) as le, sparkline(sum(n_events)) as sparkevent, sum(n_events) as events by event_input "
    '| eval \\"Last event\\" = strftime(le, \\"%e %b %Y %I:%M%p\\") ] '
    f"{FILL_DEFAULT_SPARKLINE_AND_VALUE}"
    '| makemv delim=\\",\\" sparkevent '
    '| table event_input, Active, events, sparkevent, \\"Last event\\" '
    '| rename event_input as \\"Input\\", events as \\"Number of events\\", sparkevent as \\"Event trendline\\"'
)

errors_list_query = "index=_internal source=*{addon_name}* log_level IN ({log_lvl})"

resource_cpu_query = (
    "index = _introspection component=PerProcess data.args=*{addon_name}* "
    '| timechart avg(data.pct_cpu) as \\"CPU (%)\\"'
)

resource_memory_query = (
    "index=_introspection component=PerProcess data.args=*{addon_name}* "
    '| timechart avg(data.pct_memory) as \\"Memory (%)\\"'
)


def generate_dashboard_content(
    addon_name: str,
    input_names: list[str],
    definition_json_name: str,
    lic_usg_search_params: Optional[tuple[str, str]],
    error_panel_log_lvl: str,
) -> str:
    determine_by = lic_usg_search_params[0] if lic_usg_search_params else "s"
    lic_usg_condition = (
        lic_usg_search_params[1]
        if lic_usg_search_params
        else (",".join([name + "*" for name in input_names]))
    )

    content = ""

    if definition_json_name == default_definition_json_filename["overview"]:
        content = (
            utils.get_j2_env()
            .get_template(definition_json_name)
            .render(
                data_ingestion_and_events=data_ingestion_and_events.format(
                    lic_usg_condition=lic_usg_condition,
                    addon_name=addon_name.lower(),
                    determine_by=determine_by,
                ),
                errors_count=errors_count.format(
                    addon_name=addon_name.lower(), log_lvl=error_panel_log_lvl
                ),
                errors_count_zero_line=zero_line_search_query.format(
                    value_label="Errors",
                    basic_query_token="error_count",
                    time_token="overview_time",
                ),
                data_ingestion_and_events_zero_line=zero_line_search_query.format(
                    value_label="Number of events",
                    basic_query_token="data_volume",
                    time_token="overview_time",
                ),
                events_count=events_count.format(addon_name=addon_name.lower()),
            )
        )

    if definition_json_name == default_definition_json_filename["data_ingestion_tab"]:
        content = (
            utils.get_j2_env()
            .get_template(definition_json_name)
            .render(
                data_ingestion=data_ingestion.format(
                    lic_usg_condition=lic_usg_condition, determine_by=determine_by
                ),
                data_ingestion_volume_zero_line=zero_line_search_query.format(
                    value_label="Data volume",
                    basic_query_token="data_volume",
                    time_token="data_ingestion_time",
                ),
                events_count=events_count.format(addon_name=addon_name.lower()),
                data_ingestion_event_count_zero_line=zero_line_search_query.format(
                    value_label="Number of events",
                    basic_query_token="data_ingestion_events_count",
                    time_token="data_ingestion_time",
                ),
                table_sourcetype=table_sourcetype_query.format(
                    lic_usg_condition=lic_usg_condition,
                    addon_name=addon_name.lower(),
                    determine_by=determine_by,
                ),
                table_source=table_source_query.format(
                    lic_usg_condition=lic_usg_condition,
                    addon_name=addon_name.lower(),
                    determine_by=determine_by,
                ),
                table_host=table_host_query.format(
                    lic_usg_condition=lic_usg_condition,
                    addon_name=addon_name.lower(),
                    determine_by=determine_by,
                ),
                table_index=table_index_query.format(
                    lic_usg_condition=lic_usg_condition,
                    addon_name=addon_name.lower(),
                    determine_by=determine_by,
                ),
                table_account=table_account_query.format(addon_name=addon_name.lower()),
                table_input=table_input_query.format(
                    addon_name=addon_name, addon_name_lowercase=addon_name.lower()
                ),
            )
        )

    if definition_json_name == default_definition_json_filename["errors_tab"]:
        content = (
            utils.get_j2_env()
            .get_template(definition_json_name)
            .render(
                errors_count=errors_count.format(
                    addon_name=addon_name.lower(), log_lvl=error_panel_log_lvl
                ),
                errors_count_tab_zero_line=zero_line_search_query.format(
                    value_label="Errors",
                    basic_query_token="error_count_tab",
                    time_token="errors_tab_time",
                ),
                errors_list=errors_list_query.format(
                    addon_name=addon_name.lower(), log_lvl=error_panel_log_lvl
                ),
            )
        )

    if definition_json_name == default_definition_json_filename["resources_tab"]:
        content = (
            utils.get_j2_env()
            .get_template(definition_json_name)
            .render(
                resource_cpu=resource_cpu_query.format(addon_name=addon_name.lower()),
                resource_memory=resource_memory_query.format(
                    addon_name=addon_name.lower()
                ),
            )
        )

    if (
        definition_json_name
        == default_definition_json_filename["data_ingestion_modal_definition"]
    ):
        content = (
            utils.get_j2_env()
            .get_template(definition_json_name)
            .render(
                data_ingestion=data_ingestion.format(
                    lic_usg_condition=lic_usg_condition, determine_by=determine_by
                ),
                events_count=events_count.format(addon_name=addon_name.lower()),
            )
        )

    return content


def generate_dashboard(
    global_config: global_config_lib.GlobalConfig,
    global_config_path: str,
    addon_name: str,
    definition_json_path: str,
) -> None:
    os.makedirs(os.path.normpath(definition_json_path), exist_ok=True)

    input_names = [el.get("name") for el in global_config.inputs]
    panels = global_config.dashboard.get("panels", [])
    panel_names = [panel["name"] for panel in panels]

    panels_to_display = {PANEL_DEFAULT: False, PANEL_CUSTOM: False}

    lic_usg_search_params = _get_license_usage_search_params(global_config.dashboard)

    error_panel_log_lvl = _get_error_panel_log_lvl(global_config.dashboard)

    if PANEL_DEFAULT in panel_names:
        for definition_json_name in default_definition_json_filename.values():
            content = generate_dashboard_content(
                addon_name,
                input_names,
                definition_json_name,
                lic_usg_search_params,
                error_panel_log_lvl,
            )
            with open(
                os.path.join(definition_json_path, definition_json_name), "w"
            ) as file:
                file.write(content)
        panels_to_display[PANEL_DEFAULT] = True

    if PANEL_CUSTOM in panel_names:
        dashboard_components_path = os.path.abspath(
            os.path.join(
                global_config_path,
                os.pardir,
                "custom_dashboard.json",
            )
        )
        custom_content = get_custom_json_content(dashboard_components_path)
        with open(os.path.join(definition_json_path, "custom.json"), "w") as file:
            file.write(json.dumps(custom_content))
        panels_to_display[PANEL_CUSTOM] = True

    with open(
        os.path.join(definition_json_path, "panels_to_display.json"), "w"
    ) as file:
        file.write(json.dumps(panels_to_display))


def _get_license_usage_search_params(
    dashboard: dict[Any, Any]
) -> Optional[tuple[str, str]]:
    determine_by_map = {"source": "s", "sourcetype": "st", "host": "h", "index": "idx"}

    try:
        lic_usg_type = dashboard["settings"]["custom_license_usage"]["determine_by"]
        lic_usg_search_items = dashboard["settings"]["custom_license_usage"][
            "search_condition"
        ]
    except KeyError:
        logger.info(
            "No custom license usage search condition found. Proceeding with default parameters."
        )
        return None

    determine_by = determine_by_map[lic_usg_type]
    lic_usg_condition = ",".join(['\\"' + el + '\\"' for el in lic_usg_search_items])

    return determine_by, lic_usg_condition


def _get_error_panel_log_lvl(dashboard: dict[Any, Any]) -> str:
    try:
        error_lvl = dashboard["settings"]["error_panel_log_lvl"]
    except KeyError:
        logger.info(
            "No custom error log level found. Proceeding with default parameters."
        )
        return "ERROR"
    return ", ".join(error_lvl)


def get_custom_json_content(custom_dashboard_path: str) -> dict[Any, Any]:
    custom_dashboard = load_custom_json(custom_dashboard_path)

    if not custom_dashboard:
        logger.error(
            f"Custom dashboard page set in globalConfig.json but custom content not found. "
            f"Please verify if file {custom_dashboard_path} has a proper structure "
            f"(see https://splunk.github.io/addonfactory-ucc-generator/dashboard/)"
        )
        sys.exit(1)
    return custom_dashboard


def load_custom_json(json_path: str) -> dict[Any, Any]:
    try:
        with open(json_path) as dashboard_file:
            custom_dashboard = json.load(dashboard_file)
    except FileNotFoundError:
        logger.error(
            f"Custom dashboard page set in globalConfig.json but "
            f"file {json_path} not found"
        )
        sys.exit(1)
    except json.decoder.JSONDecodeError as exc:
        logger.error(
            f"{json_path} it's not a valid json file. Error message: {str(exc)}"
        )
        sys.exit(1)
    return custom_dashboard



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import copy
from typing import Any

mapping_keys = {
    "adaptiveResponse": "adaptive_response",
    "supportsAdhoc": "supports_adhoc",
    "supportsCloud": "supports_cloud",
    "drilldownUri": "drilldown_uri",
    "entity": "parameters",
    "defaultValue": "default_value",
    "help": "help_string",
    "field": "name",
    "type": "format_type",
    "valueField": "value-field",
    "labelField": "label-field",
    "name": "short_name",
}

mapping_values = {
    "singleSelect": "dropdownlist",
    "singleSelectSplunkSearch": "dropdownlist_splunk_search",
}


def convert_list_to_dict(value_list: list[Any]) -> dict[str, Any]:
    return_list = {}
    for each_dict in value_list:
        return_list[each_dict["label"]] = each_dict["value"]
    return return_list


def transform_params(parameter_list: list[Any]) -> None:
    for param in parameter_list:
        if param["format_type"] in ["dropdownlist", "radio"]:
            options = param.pop("options")
            param["possible_values"] = convert_list_to_dict(options["items"])
        elif param["format_type"] == "dropdownlist_splunk_search":
            value_field = param.pop("value-field")
            label_field = param.pop("label-field")
            search = param.pop("search")
            param["ctrl_props"] = {
                "value-field": value_field,
                "label-field": label_field,
                "search": search,
            }
            try:
                # `options` field is optional for `dropdownlist_splunk_search`
                # type.
                options = param.pop("options")
            except KeyError:
                options = None
            if options is not None:
                earliest_time = None
                latest_time = None
                options_items = options.get("items")
                for item in options_items:
                    if item.get("label") == "earliest":
                        earliest_time = item.get("value")
                    if item.get("label") == "latest":
                        latest_time = item.get("value")
                if earliest_time is not None:
                    param["ctrl_props"]["earliest"] = earliest_time
                if latest_time is not None:
                    param["ctrl_props"]["latest"] = latest_time


def iterdict(dictionary: dict[str, Any], result: dict[str, Any]) -> None:
    """
    This function replaces key and value with the ones required by add-on alert builder
    """
    for key in dictionary:
        if key in mapping_keys:
            value = result.pop(key)
            result[mapping_keys[key]] = value
            mapped_key = mapping_keys[key]
        else:
            mapped_key = key

        if isinstance(dictionary[key], dict):
            iterdict(dictionary[key], result[mapped_key])
        elif isinstance(dictionary[key], list):
            for dictionary_item, result_item in zip(
                dictionary[key], result[mapped_key]
            ):
                if isinstance(dictionary_item, dict):
                    iterdict(dictionary_item, result_item)
        else:
            if result[mapped_key] in mapping_values:
                result[mapped_key] = mapping_values[result[mapped_key]]


def normalize(
    original_schema_content: list[dict[str, Any]], short_name: str
) -> dict[str, Any]:
    """
    Process the globalConfig alert schema to generate structure required by add-on alert generator
    """
    schema_content = {
        "alerts": original_schema_content,
    }
    result = copy.deepcopy(schema_content)
    iterdict(schema_content, result)
    for alert in result["alerts"]:
        transform_params(alert["parameters"])
    return {
        "short_name": short_name,
        "schema.content": {
            "short_name": short_name,
            "modular_alerts": result["alerts"],
        },
    }



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import ast
import json
import logging
import os
import shutil
import sys
from pathlib import Path
from typing import Optional, Any, Union
from splunk_add_on_ucc_framework.auto_gen_comparator import CodeGeneratorDiffChecker
import subprocess
import colorama as c
import fnmatch
import filecmp
import platform
import traceback

from urllib.parse import quote

from splunk_add_on_ucc_framework import (
    __version__,
    exceptions,
    global_config_update,
    global_config_validator,
    utils,
)
from splunk_add_on_ucc_framework import dashboard
from splunk_add_on_ucc_framework import meta_conf as meta_conf_lib
from splunk_add_on_ucc_framework import app_manifest as app_manifest_lib
from splunk_add_on_ucc_framework.utils import get_app_manifest
from splunk_add_on_ucc_framework import global_config as global_config_lib
from splunk_add_on_ucc_framework.commands.modular_alert_builder import (
    builder as alert_builder,
)
from splunk_add_on_ucc_framework.commands.rest_builder import (
    global_config_builder_schema,
)
from splunk_add_on_ucc_framework.commands.rest_builder.builder import RestBuilder
from splunk_add_on_ucc_framework.install_python_libraries import (
    SplunktaucclibNotFound,
    WrongSplunktaucclibVersion,
    WrongSolnlibVersion,
    install_python_libraries,
)
from splunk_add_on_ucc_framework.commands.openapi_generator import (
    ucc_to_oas,
)
from splunk_add_on_ucc_framework.generators.file_generator import begin
from splunk_add_on_ucc_framework.generators.conf_files.create_app_conf import AppConf
from splunk_add_on_ucc_framework.utils import write_file
from splunk_add_on_ucc_framework.package_files_update import handle_package_files_update

logger = logging.getLogger("ucc_gen")

internal_root_dir = os.path.dirname(os.path.dirname(__file__))


def _modify_and_replace_token_for_oauth_templates(
    ta_name: str, global_config: global_config_lib.GlobalConfig, outputdir: str
) -> None:
    """
    Rename templates with respect to addon name if OAuth is configured.

    Args:
        ta_name: Add-on name.
        global_config: Object representing globalConfig.
        outputdir: output directory.
    """
    redirect_js_src = os.path.join(
        outputdir, ta_name, "appserver", "static", "js", "build", "redirect_page.js"
    )
    redirect_html_src = os.path.join(
        outputdir, ta_name, "appserver", "templates", "redirect.html"
    )
    if global_config.has_oauth():
        html_template_path = os.path.join(outputdir, ta_name, "appserver", "templates")
        with open(os.path.join(html_template_path, "redirect.html")) as f:
            s = f.read()

        with open(os.path.join(html_template_path, "redirect.html"), "w") as f:
            s = s.replace("${ta.name}", ta_name.lower())
            s = s.replace("${ta.version}", global_config.version)
            f.write(s)

        redirect_js_dest = (
            os.path.join(outputdir, ta_name, "appserver", "static", "js", "build", "")
            + ta_name.lower()
            + "_redirect_page."
            + global_config.version
            + ".js"
        )
        redirect_html_dest = os.path.join(
            outputdir,
            ta_name,
            "appserver",
            "templates",
            ta_name.lower() + "_redirect.html",
        )
        os.rename(redirect_js_src, redirect_js_dest)
        os.rename(redirect_html_src, redirect_html_dest)
    else:
        os.remove(redirect_html_src)
        os.remove(redirect_js_src)


def _add_modular_input(
    ta_name: str,
    global_config: global_config_lib.GlobalConfig,
    outputdir: str,
    gc_path: str,
) -> None:
    for service in global_config.inputs:
        input_name = service.get("name")
        class_name = input_name.upper()
        description = service.get("title")
        entity = service.get("entity")
        field_allow_list = frozenset(["name", "interval", "index", "sourcetype"])
        template = "input.template"

        if "template" in service:
            template = service.get("template") + ".template"

        # filter fields in allow list
        entity = [x for x in entity if x.get("field") not in field_allow_list]

        input_helper_module = service.get("inputHelperModule")

        len_of_args = {"stream_events": 2, "validate_input": 1}
        include_self = {"stream_events": False, "validate_input": False}

        if input_helper_module:
            addon_root_path = os.path.dirname(gc_path)
            bin_path = os.path.join(addon_root_path, "package", "bin")
            helper_path = os.path.join(bin_path, f"{input_helper_module}.py")

            for func in len_of_args:
                try:
                    result = _get_num_of_args(func, helper_path)

                    if result is None:
                        logger.error(
                            f"Cannot find function definition 'def {func}(...)' in module '{input_helper_module}'."
                        )
                        exit(1)

                    if result == len_of_args[func]:
                        # change nothing
                        include_self[func] = False
                    elif result == len_of_args[func] + 1:
                        # add self to the function
                        include_self[func] = True
                    else:
                        logger.error(
                            f"Unexpected number of arguments in helper module. "
                            f"Function '{func}' in '{input_helper_module}' has {result} arguments."
                        )
                except FileNotFoundError:
                    # If no module, that was specified in globalConfig, default one will be created without 'self'
                    pass

        content = (
            utils.get_j2_env()
            .get_template(template)
            .render(
                input_name=input_name,
                class_name=class_name,
                description=description,
                entity=entity,
                input_helper_module=input_helper_module,
                stream_events_self=include_self["stream_events"],
                validate_input_self=include_self["validate_input"],
            )
        )
        input_file_name = os.path.join(outputdir, ta_name, "bin", input_name + ".py")
        with open(input_file_name, "w") as input_file:
            input_file.write(content)

        if input_helper_module is not None:
            helper_module_template = "input.module-template"
            helper_filename = os.path.join(
                outputdir, ta_name, "bin", f"{input_helper_module}.py"
            )

            content = (
                utils.get_j2_env()
                .get_template(helper_module_template)
                .render(
                    input_name=input_name,
                )
            )
            with open(helper_filename, "w") as helper_file:
                helper_file.write(content)


def _get_num_of_args(
    function_name: str,
    module_path: Union[str, Path],
) -> Optional[int]:
    module_path = Path(module_path)

    if not module_path.is_file():
        raise FileNotFoundError(
            f"Module path '{module_path}' does not point to a valid file."
        )

    module_content = ast.parse(module_path.read_text(encoding="utf-8"))

    for node in module_content.body:
        if isinstance(node, ast.FunctionDef) and node.name == function_name:
            # Return the number of arguments defined in the function
            return len(node.args.args)

    return None


def _get_addon_version(addon_version: Optional[str]) -> str:
    if not addon_version:
        try:
            return utils.get_version_from_git()
        except exceptions.IsNotAGitRepo:
            logger.error(
                "Could not determine a version using Git (maybe not a Git "
                "repository?). Use `--ta-version` to specify the version you "
                "want."
            )
            sys.exit(1)
        except exceptions.CouldNotVersionFromGitException:
            logger.error(
                "Could not find the proper version from git tags. "
                "Check out "
                "https://github.com/splunk/addonfactory-ucc-generator/issues/404"
            )
            sys.exit(1)
    return addon_version.strip()


def _get_build_output_path(output_directory: Optional[str] = None) -> str:
    if output_directory is None:
        return os.path.join(os.getcwd(), "output")
    else:
        if not os.path.isabs(output_directory):
            return os.path.join(os.getcwd(), output_directory)
        return output_directory


def _get_python_version_from_executable(python_binary_name: str) -> str:
    try:
        python_binary_version = subprocess.run(
            [python_binary_name, "--version"], stdout=subprocess.PIPE
        ).stdout.decode("utf-8")

        return python_binary_version.strip()
    except (subprocess.CalledProcessError, FileNotFoundError) as exc:
        raise exceptions.CouldNotIdentifyPythonVersionException(
            f"Failed to identify python version for binary {python_binary_name}. Error message: {exc}"
        )


def _get_and_check_global_config_path(source: str, config_path: Optional[str]) -> str:
    if not config_path:
        config_path = os.path.abspath(
            os.path.join(source, os.pardir, "globalConfig.json")
        )
        if not os.path.isfile(config_path):
            config_path = os.path.abspath(
                os.path.join(source, os.pardir, "globalConfig.yaml")
            )
    if os.path.isfile(config_path):
        return config_path
    return ""


def summary_report(
    source: str,
    ta_name: str,
    output_directory: str,
    verbose_file_summary_report: bool,
) -> None:
    # initialising colorama to handle ASCII color in windows cmd
    c.init()
    color_palette = {
        "copied": c.Fore.GREEN,
        "conflict": c.Fore.RED,
        "modified": c.Fore.YELLOW,
    }

    # conflicting files from ucc-gen package folder
    conflict_path = os.path.join(internal_root_dir, "package")
    # conflict files generated through-out the process
    conflict_static_list = frozenset(
        [
            "import_declare_test.py",
            f"{ta_name}_rh_*.py",
            "app.conf",
            "inputs.conf*",
            "restmap.conf",
            "server.conf",
            f"{ta_name}_*.conf*",
            "web.conf",
            "default.xml",
            "configuration.xml",
            "dashboard.xml",
            "inputs.xml",
            "openapi.json",
        ]
    )

    def line_print(print_path: str, mod_type: str) -> None:
        if verbose_file_summary_report:
            logger.info(
                color_palette.get(mod_type, "")
                + str(print_path).ljust(80)
                + mod_type
                + c.Style.RESET_ALL,
            )
        summary[mod_type] += 1

    def check_for_conflict(file: str, relative_file_path: str) -> bool:
        conflict_path_file = os.path.join(conflict_path, relative_file_path)
        if os.path.isfile(conflict_path_file):
            return True
        for pattern in conflict_static_list:
            if fnmatch.fnmatch(file, pattern):
                return True
        return False

    def file_check(
        file: str, output_directory: str, relative_file_path: str, source: str
    ) -> None:
        source_path = os.path.join(source, relative_file_path)

        if os.path.isfile(source_path):
            # file is present in package
            output_path = os.path.join(output_directory, relative_file_path)

            is_conflict = check_for_conflict(file, relative_file_path)

            if not is_conflict:
                files_are_same = filecmp.cmp(source_path, output_path)
                if not files_are_same:
                    # output file was modified
                    line_print(relative_file_path, "modified")
                else:
                    # files are the same
                    line_print(relative_file_path, "copied")
            else:
                line_print(relative_file_path, "conflict")
        else:
            # file does not exist in package
            line_print(relative_file_path, "created")

    summary = {"created": 0, "copied": 0, "modified": 0, "conflict": 0}

    path_len = len(output_directory) + 1

    if verbose_file_summary_report:
        logger.info("Detailed information about created/copied/modified/conflict files")
        logger.info(
            "Read more about it here: "
            "https://splunk.github.io/addonfactory-ucc-generator/quickstart/#verbose-mode"
        )

    for path, dir, files in os.walk(output_directory):
        relative_path = path[path_len:]
        # skipping lib directory
        if relative_path[:3] == "lib":
            if relative_path == "lib":
                line_print("lib", "created")
            continue

        files = sorted(files, key=str.casefold)

        for file in files:
            relative_file_path = os.path.join(relative_path, file)
            file_check(file, output_directory, relative_file_path, source)

    summary_combined = ", ".join(
        [
            f"{file_type}: {amount_of_files}"
            for file_type, amount_of_files in summary.items()
        ]
    )
    logger.info(f"File creation summary: {summary_combined}")


def yarn_build_custom_ui_repo(ui_dir: str, ci: bool) -> None:
    # installing dependencies
    if ci:
        subprocess.run(
            ["yarn", "--cwd", str(ui_dir), "install", "--frozen-lockfile"], check=True
        )
    else:
        subprocess.run(["yarn", "--cwd", str(ui_dir), "install"], check=True)
    # building the code
    subprocess.run(["yarn", "--cwd", str(ui_dir), "run", "build"], check=True)


def npm_build_custom_ui_repo(ui_dir: str, ci: bool) -> None:
    # installing dependencies
    if ci:
        subprocess.run(["npm", "--prefix", str(ui_dir), "ci"], check=True)
    else:
        subprocess.run(["npm", "--prefix", str(ui_dir), "install"], check=True)
    # building the code
    subprocess.run(["npm", "--prefix", str(ui_dir), "run", "build"], check=True)


def build_ui(source: str) -> None:
    ui_dir = os.path.join(source, "..", "ui")

    if not os.path.isdir(ui_dir):
        logger.error(f"UI directory does not exist {ui_dir}")
        sys.exit(1)

    # Check if node is installed
    if not shutil.which("node"):
        logger.error("Node.js is not installed. Not building custom code")
        sys.exit(1)

    try:
        ci = os.environ.get("CI", "false") == "true"

        # Check if yarn or npm is installed
        if shutil.which("yarn") and os.path.isfile(ui_dir + "/yarn.lock"):
            yarn_build_custom_ui_repo(ui_dir, ci)
        elif shutil.which("npm") and os.path.isfile(ui_dir + "/package-lock.json"):
            npm_build_custom_ui_repo(ui_dir, ci)
        else:
            logger.error(
                "Neither yarn nor npm is present with lock file. Not building custom code"
            )
            sys.exit(1)

    except subprocess.CalledProcessError as e:
        logger.error(f"Custom UI Build failed: {e}")


def exception_handler(func: Any) -> Any:
    def inner_function(*args: Any, **kwargs: Any) -> Any:
        try:
            func(*args, **kwargs)
        except Exception as e:
            system = platform.system()
            exc_type, exc_value, exc_traceback = type(e), e, e.__traceback__
            error = traceback.format_exception(exc_type, exc_value, exc_traceback)
            title = quote(f"[BUG] {str(exc_value)}")
            trb = quote("".join(error))
            url = (
                f"https://github.com/splunk/addonfactory-ucc-generator/issues/new?template=bug_report.yml&title={title}"
                f"&description={trb}&ucc_version={__version__}&system_info={system}"
            )
            logger.error(
                f"Uncaught exception occurred. Exception details:\n"
                f"{''.join(error)}\n"
                f"You can report this issue using: {url}"
            )

    return inner_function


@exception_handler
def generate(
    source: str,
    config_path: Optional[str] = None,
    addon_version: Optional[str] = None,
    output_directory: Optional[str] = None,
    python_binary_name: str = "python3",
    verbose_file_summary_report: bool = False,
    pip_version: str = "latest",
    pip_legacy_resolver: bool = False,
    pip_custom_flag: Optional[str] = None,
    build_custom_ui: bool = False,
) -> None:
    logger.info(f"ucc-gen version {__version__} is used")
    logger.info(f"Python binary name to use: {python_binary_name}")
    try:
        python_binary_version = _get_python_version_from_executable(python_binary_name)
        logger.info(f"Python Version: {python_binary_version}")
    except exceptions.CouldNotIdentifyPythonVersionException as e:
        logger.error(
            f"Failed to identify Python version for library installation. Error: {e}"
        )
        sys.exit(1)

    output_directory = _get_build_output_path(output_directory)
    logger.info(f"Output folder is {output_directory}")
    addon_version = _get_addon_version(addon_version)
    logger.info(f"Add-on will be built with version '{addon_version}'")
    if not os.path.exists(source):
        logger.error(
            f"Source directory: '{source}' does not exist. Please verify that given source exists."
        )
        sys.exit(1)
    shutil.rmtree(os.path.join(output_directory), ignore_errors=True)
    os.makedirs(os.path.join(output_directory))
    logger.info(f"Cleaned out directory {output_directory}")
    app_manifest = get_app_manifest(source)
    ta_name = app_manifest.get_addon_name()
    generated_files = []

    gc_path = _get_and_check_global_config_path(source, config_path)
    if not gc_path:
        # create one in source directory if it doesn't exist
        gc_path = os.path.sep.join(
            [os.path.abspath(os.path.dirname(source)), "globalConfig.json"]
        )
        global_config_lib.GlobalConfig.from_app_manifest(app_manifest).dump(gc_path)

    # no need of the condition as globalConfig would always exist
    logger.info(f"Using globalConfig file located @ {gc_path}")
    global_config = global_config_lib.GlobalConfig.from_file(gc_path)
    global_config.cleanup_unwanted_params()
    # handle the update of globalConfig before validating
    global_config_update.handle_global_config_update(global_config, gc_path)
    try:
        validator = global_config_validator.GlobalConfigValidator(
            internal_root_dir, global_config, source
        )
        validator.validate()
        logger.info("globalConfig file is valid")
    except exceptions.GlobalConfigValidatorException as e:
        logger.error(f"globalConfig file is not valid. Error: {e}")
        sys.exit(1)
    global_config.update_addon_version(addon_version)
    global_config.dump(gc_path)
    logger.info(
        f"Updated and saved add-on version in the globalConfig file to {addon_version}"
    )
    global_config.add_ucc_version(__version__)
    global_config.expand()
    if ta_name != global_config.product:
        logger.error(
            "Add-on name mentioned in globalConfig meta tag and that app.manifest are not same,"
            "please unify them to build the add-on."
        )
        sys.exit(1)
    scheme = global_config_builder_schema.GlobalConfigBuilderSchema(global_config)
    if global_config.has_pages():
        utils.recursive_overwrite(
            os.path.join(internal_root_dir, "package"),
            os.path.join(output_directory, ta_name),
            has_dashboard=global_config.has_dashboard(),
        )
    global_config_file = (
        "globalConfig.yaml" if gc_path.endswith(".yaml") else "globalConfig.json"
    )
    output_build_path = os.path.join(
        output_directory, ta_name, "appserver", "static", "js", "build"
    )
    if not os.path.isdir(output_build_path):
        # this path may not exist for the .conf-only add-ons
        os.makedirs(output_build_path)
    global_config.dump(os.path.join(output_build_path, global_config_file))
    logger.info("Copied globalConfig to output")
    ucc_lib_target = os.path.join(output_directory, ta_name, "lib")
    try:
        install_python_libraries(
            source,
            ucc_lib_target,
            python_binary_name,
            includes_ui=True,
            os_libraries=global_config.os_libraries,
            pip_version=pip_version,
            pip_legacy_resolver=pip_legacy_resolver,
            pip_custom_flag=pip_custom_flag,
            includes_oauth=global_config.has_oauth(),
        )
    except (
        SplunktaucclibNotFound,
        WrongSplunktaucclibVersion,
        WrongSolnlibVersion,
    ) as e:
        logger.error(str(e))
        sys.exit(1)
    logger.info(f"Installed add-on requirements into {ucc_lib_target} from {source}")
    generated_files.extend(
        begin(
            global_config=global_config, input_dir=source, output_dir=output_directory
        )
    )
    # TODO: all FILES GENERATED object: generated_files, use it for comparison
    if global_config.has_pages():
        builder_obj = RestBuilder(scheme, os.path.join(output_directory, ta_name))
        builder_obj.build()
        _modify_and_replace_token_for_oauth_templates(
            ta_name,
            global_config,
            output_directory,
        )
    if global_config.has_inputs():
        logger.info("Generating inputs code")
        _add_modular_input(ta_name, global_config, output_directory, gc_path)
    if global_config.has_alerts():
        logger.info("Generating alerts code")
        alert_builder.generate_alerts(global_config, ta_name, output_directory)

    if global_config.has_dashboard():
        logger.info("Including dashboard")
        dashboard_definition_json_path = os.path.join(
            output_directory,
            ta_name,
            "appserver",
            "static",
            "js",
            "build",
            "custom",
        )
        dashboard.generate_dashboard(
            global_config, gc_path, ta_name, dashboard_definition_json_path
        )

    utils.check_author_name(source, app_manifest)

    # Update files before overwriting
    handle_package_files_update(source)
    comparator = CodeGeneratorDiffChecker(
        source, os.path.join(output_directory, ta_name)
    )
    comparator.deduce_gen_and_custom_content(logger)
    utils.recursive_overwrite(source, os.path.join(output_directory, ta_name))
    logger.info("Copied package directory")

    default_meta_conf_path = os.path.join(
        output_directory, ta_name, "metadata", meta_conf_lib.DEFAULT_META_FILE_NAME
    )
    if not os.path.exists(default_meta_conf_path):
        os.makedirs(os.path.join(output_directory, ta_name, "metadata"), exist_ok=True)
        meta_conf = meta_conf_lib.MetaConf()
        meta_conf.write_default(default_meta_conf_path)
        logger.info(
            f"Created default {meta_conf_lib.DEFAULT_META_FILE_NAME} file in the output folder"
        )

    with open(os.path.join(output_directory, ta_name, "VERSION"), "w") as version_file:
        version_file.write(addon_version)
        version_file.write("\n")
        version_file.write(addon_version)
        logger.info("Updated VERSION file")

    app_manifest.update_addon_version(addon_version)
    output_manifest_path = os.path.abspath(
        os.path.join(output_directory, ta_name, app_manifest_lib.APP_MANIFEST_FILE_NAME)
    )
    with open(output_manifest_path, "w") as manifest_file:
        manifest_file.write(str(app_manifest))
        logger.info(
            f"Updated {app_manifest_lib.APP_MANIFEST_FILE_NAME} file in the output folder"
        )
    # NOTE: merging source and generated 'app.conf' as per previous design
    app_conf = AppConf(
        global_config=global_config, input_dir=source, output_dir=output_directory
    ).generate()
    # we need to explicitly call write method since we are not generating
    # file in child classes
    write_file(
        app_conf[0]["file_name"],
        app_conf[0]["file_path"],
        app_conf[0]["content"],
        merge_mode=app_conf[0]["merge_mode"],
    )
    license_dir = os.path.abspath(os.path.join(source, os.pardir, "LICENSES"))
    if os.path.exists(license_dir):
        logger.info("Copy LICENSES directory")
        utils.recursive_overwrite(
            license_dir, os.path.join(output_directory, ta_name, "LICENSES")
        )

    if os.path.exists(
        os.path.abspath(os.path.join(source, os.pardir, "additional_packaging.py"))
    ):
        sys.path.insert(0, os.path.abspath(os.path.join(source, os.pardir)))
        try:
            from additional_packaging import cleanup_output_files

            cleanup_output_files(output_directory, ta_name)
        except ImportError:
            logger.info(
                "additional_packaging.py is present but does not have `cleanup_output_files`. Skipping clean-up."
            )

        try:
            from additional_packaging import additional_packaging

            additional_packaging(ta_name)
        except ImportError:
            logger.info(
                "additional_packaging.py is present but does not have `additional_packaging`. "
                "Skipping additional packaging."
            )
        # clean-up sys.path manipulation
        sys.path.pop(0)

    if global_config.has_configuration() or global_config.has_inputs():
        logger.info("Generating OpenAPI file")
        open_api_object = ucc_to_oas.transform(global_config, app_manifest)

        output_openapi_folder = os.path.abspath(
            os.path.join(output_directory, ta_name, "appserver", "static")
        )
        output_openapi_path = os.path.join(output_openapi_folder, "openapi.json")
        if not os.path.isdir(output_openapi_folder):
            os.makedirs(os.path.join(output_openapi_folder))
            logger.info(f"Creating {output_openapi_folder} folder")
        with open(output_openapi_path, "w") as openapi_file:
            json.dump(open_api_object.json, openapi_file, indent=4)

    if build_custom_ui:
        logger.info("Building custom UI code")
        # should we share output dir (requires build Ui changes)
        # or should be rely on UI repo to specify it?
        build_ui(source)

    summary_report(
        source,
        ta_name,
        os.path.join(output_directory, ta_name),
        verbose_file_summary_report,
    )



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Any, Optional

from splunk_add_on_ucc_framework.entity.entity import Entity
from splunk_add_on_ucc_framework.entity.interval_entity import IntervalEntity
from splunk_add_on_ucc_framework.entity.index_entity import IndexEntity

ENTITY_TYPES = [
    IntervalEntity,
    IndexEntity,
]


def resolve_entity(
    entity_definition: dict[str, Any],
    entity_types: Optional[list[Any]] = None,
) -> Entity:
    entity_types = entity_types if entity_types else ENTITY_TYPES
    for entity_type in entity_types:
        entity = entity_type.from_definition(entity_definition)

        if entity is not None:
            return entity

    return Entity(entity_definition)


def expand_entity(entity_definition: dict[str, Any]) -> dict[str, Any]:
    return resolve_entity(entity_definition).long_form()


def collapse_entity(
    entity_definition: dict[str, Any],
    entity_types: Optional[list[Any]] = None,
) -> dict[str, Any]:
    return resolve_entity(entity_definition, entity_types).short_form()



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from splunk_add_on_ucc_framework.generators.file_generator import FileGenerator
from splunk_add_on_ucc_framework.commands.modular_alert_builder import (
    arf_consts as ac,
    normalize,
)
from typing import Any, Optional
from os import linesep
from re import search
from splunk_add_on_ucc_framework.global_config import GlobalConfig


class AlertActionsHtml(FileGenerator):
    __description__ = (
        "Generates `alert_name.html` file based on alerts configuration present in globalConfig,"
        " in `default/data/ui/alerts` folder."
    )

    def __init__(
        self, global_config: GlobalConfig, input_dir: str, output_dir: str
    ) -> None:
        super().__init__(global_config, input_dir, output_dir)
        if global_config.has_alerts():
            self._html_home = "alert_html_skeleton.template"
            envs = normalize.normalize(
                global_config.alerts,
                global_config.namespace,
            )
            schema_content = envs["schema.content"]
            self._alert_settings = schema_content["modular_alerts"]

    def generate(self) -> Optional[list[dict[str, str]]]:
        if not self._global_config.has_alerts():
            return None

        alert_details: list[Any] = []
        for self.alert in self._alert_settings:
            self.set_template_and_render(
                template_file_path=["html_templates"],
                file_name="mod_alert.html.template",
            )
            rendered_content = self._template.render(
                mod_alert=self.alert, home_page=self._html_home
            )
            text = linesep.join(
                [s for s in rendered_content.splitlines() if not search(r"^\s*$", s)]
            )
            file_name = f"{self.alert[ac.SHORT_NAME] + '.html'}"
            file_path = self.get_file_output_path(
                [
                    "default",
                    "data",
                    "ui",
                    "alerts",
                    file_name,
                ]
            )
            alert_details.append(
                {"file_name": file_name, "file_path": file_path, "content": text}
            )
        return alert_details



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import certifi
import urllib.request
import base64
import mimetypes
import os
import ssl
import uuid
import json
import logging


logger = logging.getLogger("ucc_gen")


def encode_multipart_formdata(
    fields: dict[str, str], files: dict[str, str]
) -> tuple[str, bytes]:
    """Encodes fields and files for multipart/form-data"""
    boundary = uuid.uuid4().hex
    lines = []

    for name, value in fields.items():
        lines.append(f"--{boundary}".encode())
        lines.append(f'Content-Disposition: form-data; name="{name}"'.encode())
        lines.append(b"")
        lines.append(str(value).encode("utf-8"))

    for name, filepath in files.items():
        filename = filepath.split("/")[-1]
        with open(filepath, "rb") as f:
            file_content = f.read()
        content_type = mimetypes.guess_type(filepath)[0] or "application/octet-stream"
        lines.append(f"--{boundary}".encode())
        lines.append(
            f'Content-Disposition: form-data; name="{name}"; filename="{filename}"'.encode()
        )
        lines.append(f"Content-Type: {content_type}".encode())
        lines.append(b"")
        lines.append(file_content)

    lines.append(f"--{boundary}--".encode())
    lines.append(b"")

    body = b"\r\n".join(lines)
    content_type = f"multipart/form-data; boundary={boundary}"
    return content_type, body


def upload_package(
    app_id: int,
    package_path: str,
    splunk_versions: str,
    cim_versions: str,
    visibility: bool,
    username: str,
    password: str,
) -> str:
    upload_url = f"https://splunkbase.splunk.com/api/v1/app/{app_id}/new_release/"

    fields = {
        "filename": os.path.basename(package_path),
        "cim_versions": cim_versions,
        "splunk_versions": splunk_versions,
        "visibility": str(visibility).lower(),
    }

    files = {
        "files[]": package_path,
    }

    content_type, body = encode_multipart_formdata(fields, files)
    auth_header = base64.b64encode(f"{username}:{password}".encode()).decode("utf-8")
    context = ssl.create_default_context(cafile=certifi.where())

    request = urllib.request.Request(upload_url, data=body, method="POST")
    request.add_header("Content-Type", content_type)
    request.add_header("Authorization", f"Basic {auth_header}")
    try:
        with urllib.request.urlopen(request, context=context) as response:
            response_data = response.read().decode("utf-8")
            json_data = json.loads(response_data)
            package_id = json_data.get("id", "")
            if package_id:
                logger.info(f"Package uploaded successfully. Package ID: {package_id}")
            else:
                logger.info(
                    f"Package uploaded but no package ID returned. {response_data}"
                )
            return package_id
    except urllib.error.HTTPError as e:
        error_msg = e.read().decode()
        logger.error(f"Failed to upload package. {error_msg}")
        raise


def check_package_validation(
    package_upload_id: str, username: str, password: str
) -> None:
    url = f"https://splunkbase.splunk.com/api/v1/package/{package_upload_id}/"
    auth_header = base64.b64encode(f"{username}:{password}".encode()).decode("utf-8")
    context = ssl.create_default_context(cafile=certifi.where())

    request = urllib.request.Request(url, method="GET")
    request.add_header("Authorization", f"Basic {auth_header}")

    try:
        with urllib.request.urlopen(request, context=context) as response:
            response_data = json.loads(response.read().decode("utf-8"))
            logger.info("Validation status: {}".format(response_data.get("message")))
    except urllib.error.HTTPError as e:
        error_msg = e.read().decode()
        logger.error(f"Failed to retrieve package validation status. {error_msg}")
        raise


def publish_package(
    app_id: int,
    package_path: str,
    splunk_versions: str,
    cim_versions: str,
    visibility: bool,
    username: str,
    password: str,
) -> None:
    package_upload_id = upload_package(
        app_id,
        package_path,
        splunk_versions,
        cim_versions,
        visibility,
        username,
        password,
    )
    if package_upload_id:
        check_package_validation(package_upload_id, username, password)



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
from typing import Any, Optional

from splunk_add_on_ucc_framework.tabs.tab import Tab


# defaults
NAME = "logging"
TITLE = "Logging"
LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
FIELD = "loglevel"
LABEL = "Log level"
DEFAULTLEVEL = "INFO"

ENTITY_KEYS_REQUIRED = {"type", "label", "options", "field"}
ENTITY_KEYS_OPTIONAL = {"help", "defaultValue", "required", "validators"}
AVAILABLE_LEVELS = {"DEBUG", "INFO", "WARN", "WARNING", "ERROR", "CRITICAL"}


class LoggingTab(Tab):
    @property
    def tab_type(self) -> Optional[str]:
        return "loggingTab"

    def short_form(self) -> dict[str, Any]:
        entity = self["entity"][0]
        levels = [i["value"] for i in entity["options"]["autoCompleteFields"]]
        new_definition = {"type": "loggingTab"}

        for key, value, default in [
            ("name", self["name"], NAME),
            ("title", self["title"], TITLE),
            ("label", entity["label"], LABEL),
            ("field", entity["field"], FIELD),
            ("levels", levels, LEVELS),
            ("defaultLevel", entity.get("defaultValue", DEFAULTLEVEL), DEFAULTLEVEL),
        ]:
            if value != default:
                new_definition[key] = value

        if "help" in entity:
            new_definition["help"] = entity["help"]

        return new_definition

    @classmethod
    def from_definition(cls, definition: dict[str, Any]) -> Optional["Tab"]:
        """
        This function checks if the definition either has type==loggingTab, or if it is a normal tab (which later will
        be converted), that satisfies the following conditions:
        1. This dictionary has only 3 keys: name, title and entity (e.g. no keys like warnings)
        2. It has only one entity
        3. The entity is singleSelect and has the predefined log levels.

        Note: Although it is possible to set custom levels, this function will omit other levels, as it would be harder
        to determine whether the tab is indeed a logging tab.
        """
        if definition.get("type") == "loggingTab":
            levels = definition.get("levels", LEVELS)

            entity = {
                "type": "singleSelect",
                "label": definition.get("label", LABEL),
                "options": {
                    "disableSearch": True,
                    "autoCompleteFields": [
                        {"value": lvl, "label": lvl} for lvl in levels
                    ],
                },
                "defaultValue": definition.get("defaultLevel", DEFAULTLEVEL),
                "field": definition.get("field", FIELD),
                "required": True,
                "validators": [
                    {
                        "type": "regex",
                        "errorMsg": f"Log level must be one of: {', '.join(levels)}",
                        "pattern": f"^{'|'.join(re.escape(lvl) for lvl in levels)}$",
                    }
                ],
            }

            new_definition = {
                "name": definition.get("name", NAME),
                "title": definition.get("title", TITLE),
                "entity": [entity],
            }

            if "help" in definition:
                entity["help"] = definition["help"]

            return LoggingTab(new_definition)

        if "type" in definition:
            return None

        if definition.keys() != {"name", "title", "entity"}:
            return None

        if len(definition["entity"]) != 1:
            return None

        entity = definition["entity"][0]

        if not all(key in entity.keys() for key in ENTITY_KEYS_REQUIRED):
            return None

        if entity.keys() - ENTITY_KEYS_REQUIRED - ENTITY_KEYS_OPTIONAL:
            return None

        if entity["type"] != "singleSelect":
            return None

        if entity["options"].keys() != {
            "disableSearch",
            "autoCompleteFields",
        } and entity["options"].keys() != {"autoCompleteFields"}:
            return None

        levels = []

        for field in entity["options"]["autoCompleteFields"]:
            if "value" not in field:
                return None

            level = field["value"]

            if level not in AVAILABLE_LEVELS:
                return None

            levels.append(level)

        entity["required"] = True
        entity["options"]["disableSearch"] = True

        return LoggingTab(definition)
